// BeginLatex
// The focus of this example is to demonstrate ways to improve the robustness 
// of the registration process. Basic concept on feature-based registration
// is explained in \verb+registration1.cxx+, which should be fully understood 
// before proceeding to this example. 
// 
// The following header files provide 
// declarations for classes required for robust registration. 
//
// EndLatex


#include <vcl_fstream.h>
#include <vcl_iostream.h>
#include <vnl/vnl_vector_fixed.h>

#include <rrel/rrel_muset_obj.h>
#include <rrel/rrel_tukey_obj.h>

#include <rgrl/rgrl_feature_based_registration.h>
#include <rgrl/rgrl_feature_set_location_masked.h>
#include <rgrl/rgrl_initializer_prior.h>
#include <rgrl/rgrl_matcher_k_nearest.h>
#include <rgrl/rgrl_trans_quadratic.h>
#include <rgrl/rgrl_est_quadratic.h>

// BeginCodeSnippet
#include <rgrl/rgrl_feature_trace_pt.h>
#include <rgrl/rgrl_scale_est_all_weights.h>
#include <rgrl/rgrl_scale_est_closest.h>
#include <rgrl/rgrl_weighter_m_est.h>
#include <rgrl/rgrl_convergence_on_weighted_error.h>
// EndCodeSnippet

#include <rgrl/rgrl_roi.h>
#include <rgrl/rgrl_cast.h>
#include <rgrl/rgrl_mask.h>
#include <rgrl/rgrl_converge_status.h>
#include <rgrl/rgrl_data_manager.h>

typedef vcl_vector< rgrl_feature_sptr >  feature_vector;
typedef vnl_vector_fixed<double,2>       vector_2d;

// BeginLatex
//
// In this example, we want to register 2 retinal images using features
// extracted from the vessels. Information available from the features includes
// the center point locations and the tangent directions of the vessels at 
// these points. The transformation model is quadratic (12 parameters).
// The moving image is registered to the fixed image starting with
// an \emph{identity} transform. In this section, we'll only discuss components
// that are different from those in \verb+registration1.cxx+ for robust
// registration.
//
// In \verb+registration1.cxx+, we perform the registration based on
// point-to-point matching. To make the process less sensitive to
// sampling of feature points, point-to-line matching is a more robust
// choice (See Figure~\ref{fig:distance}). To achieve point-to-line
// matching, each feature from the fixed image stores both the center
// location and the tangent direction of the centerline. Thus, each
// feature becomes a linear approximation of the centerline at the
// feature center location.  Such feature is called a trace point
// (\code{rgrl\_feature\_trace\_pt}). For a point-to-point match, the
// error distance between the transformed point and its match is the
// Euclidean distance. For a point-to-line match, the error metric is
// the shortest distance between the transformed point and the
// matching line.
//
// The function below reads in the
// trace points. For simplicity, both feature sets contain trace
// points.
// 
// EndLatex

// BeginCodeSnippet
void 
read_feature_file( const char* filename,
                   feature_vector& trace_points )
{
// EndCodeSnippet
  vcl_ifstream istr( filename );

  if( !istr ) {
    vcl_cerr<<"ERROR: Cannot open "<<filename<<vcl_endl;
    return;
  }

  vector_2d location;
  vector_2d direction;

  // BeginLatex
  // 
  // \verb+...+
  //
  //EndLatex

  // BeginCodeSnippet
  bool done = false;
  while ( !done && istr ) {
    if ( !(istr >> location[0] >> location[1] >> direction[0] >> direction[1]) )
      done = true;
    else trace_points.push_back( new rgrl_feature_trace_pt(location, direction) );
  }
  // EndCodeSnippet

  istr.close();
  vcl_cout<<"There are "<<trace_points.size()<<" features"<<vcl_endl;

//BeginCodeSnippet
}
//EndCodeSnippet

// BeginLatex
//
// \begin{figure}[tbp]
// \centerline{\includegraphics[width=5in]{distance_metric_new}}
// \caption{A schematic illustrating trace point matching and the
// distance metric.  Trace point $\vect{p}$ from $I_1$ (the moving
// image) is mapped into $I_2$ (the fixed image) based on the current
// transformation estimate (point $\vect{p}'$), and the closest trace
// $\vect{q}$ in $I_2$ is found.  (The small arrows represent the
// trace directions.)  The error distance is the distance from
// $\vect{p}'$ to the line tangent to the trace through $\vect{q}$
// (the dotted line).  Letting $\vecg{\eta}_q$ be the normal to the
// line, the distance is $|(\vect{p}'-\vect{q})^T\vecg{\eta}_q|$.}
// \label{fig:distance} \end{figure}
//
// EndLatex

extern void (*fb_callback)( rgrl_transformation_sptr trans, rgrl_set_of<rgrl_match_set_sptr> const& match_sets );
void callback( rgrl_transformation_sptr trans, rgrl_set_of<rgrl_match_set_sptr> const& match_sets )
{
  rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
  vcl_cout<<"xform: Q \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
    "t = "<<q_xform->t()<<vcl_endl;
 }

int 
main( int argc, char* argv[] )
{

  if( argc < 3 ) {
    vcl_cerr << "Missing Parameters " << vcl_endl;
    vcl_cerr << "Usage: " << argv[0];
    vcl_cerr << " FixedImageFeatureFile MovingImageFeatureFile";
    return 1;
  }

  // BeginLatex
  //
  // In the main body of the program, we first read in the features
  // from the external files. 
  //
  // EndLatex

  feature_vector moving_feature_points;
  feature_vector fixed_feature_points;
  const char* fixed_file_name = argv[1];
  const char* moving_file_name = argv[2];

  // BeginCodeSnippet
  read_feature_file( moving_file_name, moving_feature_points );
  read_feature_file( fixed_file_name, fixed_feature_points );
  // EndCodeSnippet

  // BeginLatex
  //
  // Next, we prepare the feature sets. Just to complicate the example,
  // we assume that we're only interested in mapping a small region of the
  // moving image to the fixed. To achieve this, we first define the mask for
  // the region of interest.
  //
  // EndLatex

  // BeginCodeSnippet
  vector_2d roi_x0(200, 100);
  vector_2d roi_x1(1000, 900);
  rgrl_mask_sptr mask = new rgrl_mask_box( roi_x0, roi_x1 );
  // EndCodeSnippet

  // BeginLatex
  //
  // A masked feature set, \code{rgrl\_feature\_set\_location\_masked}, 
  // differs from its super-class, \code{rgrl\_feature\_set\_location}, by 
  // checking if the requested features are in the valid region.     
  //
  // EndLatex
  
  // BeginCodeSnippet
  const unsigned int dimension = 2;
  rgrl_feature_set_sptr moving_feature_set;
  rgrl_feature_set_sptr fixed_feature_set;
  moving_feature_set = 
    new rgrl_feature_set_location_masked<dimension>(moving_feature_points,
                                                    mask);
  fixed_feature_set = 
    new rgrl_feature_set_location<dimension>(fixed_feature_points);
  // EndCodeSnippet

  
  // Transformation estimator
  //
  int dof = 12;                    //parameter degree of freedom
  int numSampleForFit = 12;        //minimum number of samples for a fit
  rgrl_estimator_sptr estimator = new rgrl_est_quadratic(dof, 
                                                         numSampleForFit);

  //set the initial transformation to identity
  //
  rgrl_transformation_sptr init_trans; 
  vnl_matrix<double> Q(2,3,0.0);
  vnl_matrix<double> A(2,2,vnl_matrix_identity);
  vector_2d          t(0.0, 0.0);
  vnl_matrix<double> covariance_matrix( dof, dof, 0.0 );
  init_trans = new rgrl_trans_quadratic(Q, A, t, covariance_matrix);

  //Initializer
  vector_2d x0(0,0);           //upper left corner
  vector_2d x1(1023,1023);     //bottom right corner
  rgrl_roi moving_image_region(x0, x1);
  rgrl_roi fixed_image_region(x0, x1);
  rgrl_initializer_sptr initializer = 
    new rgrl_initializer_prior( moving_image_region,  
                                fixed_image_region, 
                                estimator, 
                                init_trans);

  //Matcher
  //
  unsigned int k = 1;
  rgrl_matcher_sptr cp_matcher = new rgrl_matcher_k_nearest( k );

  // BeginLatex
  //
  // To robustly handle outliers (wrong matches) generated by the
  // matcher, proper weights should be assigned to the matches, such
  // that matches with higher errors have less weight. The weight
  // function depends on the loss function for the estimation of the
  // transformation \cite{stewart:siamrev99}.
  //
  // The choice of loss function, \code{m\_est\_obj}, is motivated by
  // looking at several different loss functions and their associated
  // weight functions illustrated in Figure~\ref{fig:robust-loss}.
  // The least-squares loss function has a constant weight (unit
  // weight), the Cauchy weight function descends and asymptotes at 0,
  // while the Beaton-Tukey biweight function has a hard limit beyond
  // which the weight is 0.  This limit is set to about $4 \hat
  // \sigma$ (error scale) in the statistics literature
  // \cite{holland:stats77}.  This is important for rejecting errors
  // due to mismatches, and therefore it is the one we choose.
  //
  // EndLatex

  // BeginCodeSnippet
  vcl_auto_ptr<rrel_m_est_obj>  m_est_obj( new rrel_tukey_obj(4) );
  rgrl_weighter_sptr wgter = new rgrl_weighter_m_est(m_est_obj, false, false);
  // EndCodeSnippet

  // BeginLatex
  //  
  // \begin{figure}[tbp]
  // \begin{center}
  // \begin{tabular}{cc}
  // \includegraphics[width=2.3in]{m_est_rho} &
  // \includegraphics[width=2.3in]{m_est_wgts} \\
  // (a) & (b) 
  // \end{tabular}
  // \end{center}
  // \caption{Plots of (a) the robust loss function $\rho(u)$ and (b) weight
  // function $w(u) = \rho'(u) / u $ for the Beaton-Tukey biweight loss
  // function, the Cauchy loss function and the quadratic loss function,
  // which equates to least-squares estimation.  The Beaton-Tukey is chosen
  // because it most aggressively rejects outliers, providing no weight to
  // matches with normalized distances greater than about 4 standard
  // deviations.}
  // \label{fig:robust-loss}
  // \end{figure}
  //
  // EndLatex

  // BeginLatex
  //
  // It is clear that the weight of a match depends on both the loss
  // function and the registration error of the corresponding
  // features.  The error should be measured in error scale ($\hat
  // \sigma$). There are several ways to compute the scale, and the
  // methods can be classified into weighted and unweighted. We use
  // both in this example. Weighted scale estimator has precedence
  // over unweighted, but it requires initialization (to get the first
  // set of weights). This can be done either by providing a prior
  // scale to the initializer, or an unweighted scale estimator. We
  // take the second approach in the example. The unweighted scale
  // estimator uses MUSE \cite{miller:cvpr96}, which is similar to MAD
  // (Median Absolute Deviation), but it automatically determines the
  // fraction of (approximately) correct matches.
  // 
  // EndLatex

  // BeginCodeSnippet
  int max_set_size = 3000;  //maximum expected number of features
  vcl_auto_ptr<rrel_objective> muset_obj( new rrel_muset_obj( max_set_size , false) );

  rgrl_scale_estimator_unwgted_sptr unwgted_scale_est;
  rgrl_scale_estimator_wgted_sptr wgted_scale_est;

  unwgted_scale_est = new rgrl_scale_est_closest( muset_obj );
  wgted_scale_est = new rgrl_scale_est_all_weights();
  // EndCodeSnippet

  // BeginLatex
  //
  // To determine the convergence, we use the robust weights to
  // compute the average weighted error. If the weights are correctly
  // assigned, weighted error is more robust than the median error,
  // and can often handle more than 50\% outliers.
  //
  // EndLatex

  // BeginCodeSnippet
  double tolerance = 1.5;
  rgrl_convergence_tester_sptr conv_test = 
    new rgrl_convergence_on_weighted_error( tolerance );
  // EndCodeSnippet

  // BeginLatex
  //
  // We now have all the components and data ready for the registration 
  // process. We pass in the data sets, the matcher, the weighter, 
  // the scale estimators, and the transformation model for the $0^{th}$ 
  // stage. 
  //
  // EndLatex

  fb_callback = callback;

  // BeginCodeSnippet
  rgrl_data_manager_sptr data = new rgrl_data_manager();
  data->add_data( moving_feature_set,   // data from moving image
                  fixed_feature_set,    // data from fixed image
                  cp_matcher,           // matcher for this data
                  wgter,                // weighter
                  unwgted_scale_est,    // unweighted scale estimator
                  wgted_scale_est );    // weighted scale estimator

  rgrl_feature_based_registration reg( data, conv_test );


  reg.run( initializer );
  // EndCodeSnippet

  // Output Results
  if ( reg.has_final_transformation() ) {
    vcl_cout<<"Final xform: "<<vcl_endl;
    rgrl_transformation_sptr trans = reg.final_transformation();
    rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
    vcl_cout<<"Initial xform: Q \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
      "t = "<<q_xform->t()<<vcl_endl;
    vcl_cout<<"Final alignment error = "<<reg.final_status()->error()<<vcl_endl;
  }

  // BeginLatex
  //
  // Let's execute this example with the feature files 
  // \verb+rgrl/data/IMG0002.pgm.txt+ and \verb+rgrl/data/IMG0003.pgm.txt+
  //
  // EndLatex
  
  return 0;
}
