// BeginLatex
//
// The focus of this example is to demonstrate multiresolution
// feature-based registration. The advantage of such an approach is
// the ability to escape numerous local minima at the lower
// resolution.  Different models can also be used at different
// resolutions for better stability and wider domain of convergence
// \cite{bergen:eccv92,feldmar:cviu97,sawhney:pami99}.
// 
// In this example, we register the retinal images using 2
// resolutions: the original resolution and the down-sampled. The
// image of the lower resolution is half the size for each dimension
// of the original. We also assign different transformation models to
// different resolutions. An affine transformation model (6
// parameters) is for the lower resolution and a quadratic model (12
// parameters) is for the original. Starting at the lower resolution,
// the registration process is initialized with an \emph{identity}
// transform.  In this section, we'll only discuss steps that are
// different from \verb+registration2.cxx+ for multiresolution
// registration.
//
// EndLatex


#include <vcl_fstream.h>
#include <vcl_iostream.h>
#include <vnl/vnl_vector_fixed.h>

#include <rrel/rrel_muset_obj.h>
#include <rrel/rrel_tukey_obj.h>

#include <rgrl/rgrl_feature_based_registration.h>
#include <rgrl/rgrl_feature_trace_pt.h>
#include <rgrl/rgrl_feature_set_location.h>
#include <rgrl/rgrl_initializer_prior.h>
#include <rgrl/rgrl_matcher_k_nearest.h>
#include <rgrl/rgrl_est_quadratic.h>
#include <rgrl/rgrl_trans_quadratic.h>
#include <rgrl/rgrl_est_affine.h>
#include <rgrl/rgrl_trans_affine.h>

#include <rgrl/rgrl_weighter_m_est.h>
#include <rgrl/rgrl_convergence_on_weighted_error.h>
#include <rgrl/rgrl_scale_est_all_weights.h>
#include <rgrl/rgrl_scale_est_closest.h>

#include <rgrl/rgrl_roi.h>
#include <rgrl/rgrl_cast.h>
#include <rgrl/rgrl_converge_status.h>
#include <rgrl/rgrl_data_manager.h>


typedef vcl_vector< rgrl_feature_sptr >  feature_vector;
typedef vnl_vector_fixed<double,2>       vector_2d;

void 
read_feature_file( const char*     filename,
                   feature_vector& trace_points )
{
  vcl_ifstream istr( filename );

  if( !istr ) {
    vcl_cerr<<"ERROR: Cannot open "<<filename<<vcl_endl;
    return;
  }

  vector_2d location;
  vector_2d direction;
  bool done = false;
  while ( !done && istr ) {
    if ( !(istr >> location[0] >> location[1] >> direction[0] >> direction[1]) )
      done = true;
    else trace_points.push_back( new rgrl_feature_trace_pt(location, direction) );
  }
  istr.close();
  vcl_cout<<"There are "<<trace_points.size()<<" features"<<vcl_endl;
}

extern void (*fb_callback)( rgrl_transformation_sptr trans, rgrl_set_of<rgrl_match_set_sptr> const& match_sets );
void callback( rgrl_transformation_sptr trans, rgrl_set_of<rgrl_match_set_sptr> const& match_sets  )
{  
  if ( trans->type() == RGRL_AFFINE_XFORM ) {
    rgrl_trans_affine* a_xform = rgrl_cast<rgrl_trans_affine*>(trans);
    vcl_cout<<"xform: A = \n"<<a_xform->A()<< "t = "<<a_xform->t()<<vcl_endl;
  }
  else if ( trans->type() == RGRL_QUADRATIC_XFORM ) {
    rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
    vcl_cout<<"xform: Q = \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
      "t = "<<q_xform->t()<<vcl_endl;
  }
}

int 
main( int argc, char* argv[] )
{
  if( argc < 5 ) {
    vcl_cerr << "Missing Parameters " << vcl_endl;
    vcl_cerr << "Usage: " << argv[0];
    vcl_cerr << " FixedImageFeatureFileHighRes FixedImageFeatureFileLowRes MovingImageFeatureFileHighRes MovingImageFeatureFileLowRes";
    return 1;
  }

  //read in the feature files
  //
  feature_vector moving_fps_high_res;
  feature_vector moving_fps_low_res;
  feature_vector fixed_fps_high_res;
  feature_vector fixed_fps_low_res;

  const char* fixed_file_name_high_res = argv[1];
  const char* fixed_file_name_low_res = argv[2];
  const char* moving_file_name_high_res = argv[3];
  const char* moving_file_name_low_res = argv[4];

  read_feature_file( moving_file_name_high_res, moving_fps_high_res );
  read_feature_file( moving_file_name_low_res, moving_fps_low_res );
  read_feature_file( fixed_file_name_high_res, fixed_fps_high_res );
  read_feature_file( fixed_file_name_low_res, fixed_fps_low_res );

  // Prepare feature sets
  //
  rgrl_feature_set_sptr moving_feature_set_high_res = new rgrl_feature_set_location<2>(moving_fps_high_res);
  rgrl_feature_set_sptr moving_feature_set_low_res = new rgrl_feature_set_location<2>(moving_fps_low_res);
  rgrl_feature_set_sptr fixed_feature_set_high_res = new rgrl_feature_set_location<2>(fixed_fps_high_res);
  rgrl_feature_set_sptr fixed_feature_set_low_res = new rgrl_feature_set_location<2>(fixed_fps_low_res);

  // Set the initial transformation to be identity
  //
  int dof = 6;                      //dof for affine model
  rgrl_transformation_sptr init_trans; 
  vnl_matrix<double> A(2,2,vnl_matrix_identity);
  vector_2d          t(0.0, 0.0);
  vnl_matrix<double> covariance_matrix( dof, dof, 0.0 );

  init_trans = new rgrl_trans_affine(A, t, covariance_matrix);

  //set the component for the registration
  //
  fb_callback = callback;

  // BeginLatex
  //
  // We first instantiate the 2 transformation models, and the
  // initializer for the lower resolution. Please note, the
  // initializer has to have the knowledge of the resolution level at
  // which the registration begins. The higher the number is for the
  // resolution level, the lower the image resolution. The highest
  // resolution is always at the $0^{th}$ level.
  //
  //EndLatex
  
  // BeginCodeSnippet
  rgrl_estimator_sptr affine_model = new rgrl_est_affine(6, 6);
  rgrl_estimator_sptr quadratic_model = new rgrl_est_quadratic(12, 12);

  vector_2d x0(0,0);
  vector_2d x1(511, 511);
  rgrl_roi moving_image_region(x0, x1);
  rgrl_roi fixed_image_region(x0, x1);
  int initial_resolution = 1;
  rgrl_initializer_sptr initializer = 
    new rgrl_initializer_prior(moving_image_region, 
                               fixed_image_region, 
                               affine_model, 
                               init_trans, 
                               initial_resolution );
  // EndCodeSnippet

  // Create matcher
  //
  unsigned int k = 1;
  rgrl_matcher_sptr cp_matcher = new rgrl_matcher_k_nearest( k );

  // Robust weighting
  //
  vcl_auto_ptr<rrel_m_est_obj>  m_est_obj( new rrel_tukey_obj(4) );
  rgrl_weighter_sptr wgter = new rgrl_weighter_m_est(m_est_obj, false, false);

  // Scale estimator both weighted and unweighted
  //
  int max_set_size = 3000;
  vcl_auto_ptr<rrel_objective> muset_obj( new rrel_muset_obj( max_set_size , false) );

  rgrl_scale_estimator_unwgted_sptr unwgted_scale_est;
  rgrl_scale_estimator_wgted_sptr wgted_scale_est;

  unwgted_scale_est = new rgrl_scale_est_closest( muset_obj );
  wgted_scale_est = new rgrl_scale_est_all_weights();
 
  // Convergence tester
  //
  double tolerance = 1.5;
  rgrl_convergence_tester_sptr conv_test = 
    new rgrl_convergence_on_weighted_error( tolerance );

  //Set up the initialization problem
  //

  // BeginLatex
  //
  // We add data and appropriate models to both stages the same way as
  // we have done in the previous 2 examples. The only new step here
  // is to specify the transition going from the lower to the higher
  // resolution using
  // \code{set\_dimension\_increase\_for\_next\_stage()} method. This
  // method sets the parameter that allows the internal data to be
  // scaled properly when moving to the next resolution level. In this
  // example, since the lower resolution image is half the size for
  // each dimension of the original, the \code{dimension\_increase} is
  // set to 2 for the lower resolution.
  //
  // EndLatex

  // BeginCodeSnippet
  bool multi_resol = true;
  rgrl_data_manager_sptr data = new rgrl_data_manager( multi_resol );

  int resolution = 0;              //original resolution
  data->add_data( resolution,                 
                  moving_feature_set_high_res,
                  fixed_feature_set_high_res, 
                  cp_matcher, 
                  wgter,           
                  unwgted_scale_est,    
                  wgted_scale_est );              
  data->add_estimator( resolution, quadratic_model);

  resolution = 1;                   //lower resolution
  double dimension_increase = 2;
  data->add_data( resolution,
                  moving_feature_set_low_res,
                  fixed_feature_set_low_res, 
                  cp_matcher,
                  wgter,           
                  unwgted_scale_est,    
                  wgted_scale_est );              
  data->add_estimator( resolution, affine_model);
  data->set_dimension_increase_for_next_stage( resolution, dimension_increase);
  //EndCodeSnippet
  
  // Start the registration process
  //
  rgrl_feature_based_registration reg( data, conv_test );
  reg.run( initializer );

  // Output Results
  //
  if ( reg.has_final_transformation() ) {
    vcl_cout<<"Final xform: "<<vcl_endl;
    rgrl_transformation_sptr trans = reg.final_transformation();
    rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
    vcl_cout<<"Initial xform: Q \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
      "t = "<<q_xform->t()<<vcl_endl;
    vcl_cout<<"Final alignment error = "<<reg.final_status()->error()<<vcl_endl;
  }

  // BeginLatex
  //
  // Let's execute this example again with the feature files
  // \verb+rgrl/data/IMG0002.pgm.txt+ and
  // \verb+rgrl/data/IMG0003.pgm.txt+. The corresponding low resolution
  // files are \verb+rgrl/data/IMG0002_low.pgm.txt+ and
  // \verb+rgrl/data/IMG0003_low.pgm.txt+, respectively.
  //
  // EndLatex
  
  return 0;
}
