// BeginLatex
//
// This example is a simplified version of the hierarchical pairwise
// registration algorithm in \cite{can:pami-pair-02}. This algorithm
// combines the use of landmarks, multiple matches enhanced by
// signature similarity, and RANSAC \cite{fisch:cacm81}
// initialization.
//
// Different from the previous examples, the registration features are
// landmarks. So, what is a landmark? A landmark feature is defined as
// a branching or cross-over point of a vascular-like structure. It is
// characterized by a number of outgoing direction vectors. To make it
// more general, any feature with a center location and a set of
// outgoing direction vectors can be a "landmark", such as a corner.
//
// For each match of landmarks, the associated similarity/signature
// weight is a measure of the best alignment of the 2 sets of outgoing
// directions. It is computed as the sum of the dot-product of the
// matched directions, normalized by twice the size of the smaller
// set of outgoing directions. The robust weight based on the
// Euclidean distance is augmented by the similarity weight. Please
// refer to \cite{can:pami-pair-02} for details.
//
// The simplified hierarchical pairwise registration algorithm consists 
// of following steps: 
// \begin{enumerate} 
// \item Manually guess the initial translation (need not be very accurate). 
// \item Based on the initial translation, generate 2 matches for each 
// landmark in the moving image. Distance threshold maybe applied to 
// further reduce the number of matches.
// \item Run random sampling on the match set to obtain the best affine 
// transformation, which serves as the initial transformation for IRLS. 
// \item Run IRLS with the match set to estimate the final quadratic 
// transformation. 
// \end{enumerate}
//
// EndLatex


#include <vcl_fstream.h>
#include <vcl_iostream.h>
#include <vnl/vnl_vector_fixed.h>

#include <vil/vil_load.h>
#include <vil/vil_image_view.h>

#include <rrel/rrel_muset_obj.h>
#include <rrel/rrel_tukey_obj.h>

#include <rgrl/rgrl_feature_based_registration.h>
#include <rgrl/rgrl_initializer_ran_sam.h>
#include <rgrl/rgrl_feature_landmark.h>
#include <rgrl/rgrl_feature_set_location_masked.h>
#include <rgrl/rgrl_mask.h>
#include <rgrl/rgrl_matcher_k_nearest.h>
#include <rgrl/rgrl_match_set.h>

#include <rgrl/rgrl_est_affine.h>
#include <rgrl/rgrl_est_quadratic.h>
#include <rgrl/rgrl_trans_translation.h>
#include <rgrl/rgrl_trans_affine.h>
#include <rgrl/rgrl_trans_quadratic.h>

#include <rgrl/rgrl_scale.h>
#include <rgrl/rgrl_scale_est_all_weights.h>
#include <rgrl/rgrl_scale_est_closest.h>
#include <rgrl/rgrl_weighter_m_est.h>
#include <rgrl/rgrl_convergence_on_weighted_error.h>

#include <rgrl/rgrl_roi.h>
#include <rgrl/rgrl_cast.h>
#include <rgrl/rgrl_mask.h>
#include <rgrl/rgrl_converge_status.h>
#include <rgrl/rgrl_data_manager.h>

typedef vcl_vector< rgrl_feature_sptr >  feature_vector;
typedef vnl_vector_fixed<double,2>       vector_2d;
typedef vcl_vector< vnl_vector<double> > vec_vec_type;

void 
read_feature_file( const char* filename,
                   feature_vector& landmarks )
{
  vcl_ifstream istr( filename );

  if( !istr ) {
    vcl_cerr<<"ERROR: Cannot open "<<filename<<vcl_endl;
    return;
  }

  vector_2d location;
  vector_2d direction;
  int num_landmarks;
  int num_directions;
  double temp1, temp2; //to store information not needed in this example

  istr >> num_landmarks;
  for ( unsigned int li = 0; li < num_landmarks; li++ ) {
    istr >> location[0] >> location[1] >> temp1 >> temp2 >> num_directions;
    vec_vec_type directions;
    for ( unsigned int di = 0; di < num_directions; di++ ) {
      istr >> direction[0] >> direction[1];
      directions.push_back( direction );
    }
    landmarks.push_back( new rgrl_feature_landmark(location, directions) );
  }

  istr.close();
  vcl_cout<<"There are "<<landmarks.size()<<" landmarks"<<vcl_endl;

}

extern void (*fb_callback)( rgrl_transformation_sptr trans, rgrl_set_of<rgrl_match_set_sptr> const& match_sets );
void callback( rgrl_transformation_sptr trans, rgrl_set_of<rgrl_match_set_sptr> const& match_sets )
{
  if ( trans->type() == RGRL_AFFINE_XFORM ) {
    rgrl_trans_affine* affine_xform = rgrl_cast<rgrl_trans_affine*>(trans);
    vcl_cout<<"Initial xform: A \n"<<affine_xform->A()<<"\nt = "<<affine_xform->t()<<vcl_endl;
  }
  else if ( trans->type() == RGRL_QUADRATIC_XFORM ) {
    rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
    vcl_cout<<"xform: Q \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
      "t = "<<q_xform->t()<<vcl_endl;
  }
  else vcl_cout<<"Unkown type"<<vcl_endl;
}

int 
main( int argc, char* argv[] )
{

  if( argc < 4 ) {
    vcl_cerr << "Missing Parameters " << vcl_endl;
    vcl_cerr << "Usage: " << argv[0];
    vcl_cerr << " FixedImageLandmarkFile MovingImageLandmarkFile MaskImage";
    return 1;
  }

  // BeginLatex
  //
  // In the main body of the program, we first read in the features
  // from the external files. 
  //
  // EndLatex

  feature_vector moving_landmark_set;
  feature_vector fixed_landmark_set;
  const char* fixed_file_name = argv[1];
  const char* moving_file_name = argv[2];

  // BeginCodeSnippet
  read_feature_file( moving_file_name, moving_landmark_set );
  read_feature_file( fixed_file_name, fixed_landmark_set );
  // EndCodeSnippet

  // BeginLatex
  //
  // Next, we prepare the feature sets. Different from
  // \verb+registration2.cxx+, the mask is defined by an image with
  // non-zero intensity value for valid pixel positions.  
  //
  // EndLatex

  const unsigned int dimension = 2;
  rgrl_feature_set_sptr moving_feature_set;
  rgrl_feature_set_sptr fixed_feature_set;

  // BeginCodeSnippet

  const char* make_file_name = argv[3];
  vil_image_view<vxl_byte> mask_image;
  mask_image = vil_load(make_file_name);
  rgrl_mask_sptr mask = new rgrl_mask_2d_image( mask_image );

  moving_feature_set = 
    new rgrl_feature_set_location_masked<dimension>(moving_landmark_set,
                                                    mask);
  fixed_feature_set = 
    new rgrl_feature_set_location_masked<dimension>(fixed_landmark_set,
                                                    mask);
  // EndCodeSnippet

  vector_2d x0(0,0);           //upper left corner
  vector_2d x1(1023,1023);     //bottom right corner
  rgrl_roi moving_image_region(x0, x1);
  rgrl_roi fixed_image_region(x0, x1);
  
  // BeginLatex
  //
  // Now we're ready to generate the matches and hierarchically
  // estiamte the transformation. We first manually specify the
  // initial translation, and then generate the k-nearest matches,
  // where $k=2$ in this example.
  //
  // EndLatex

  // BeginCodeSnippet

  double shift_x = -286;
  double shift_y = -42;
  vector_2d trans( shift_x, shift_y); 
  vnl_matrix<double> covariance_matrix( 2, 2, 0.0 );
  rgrl_transformation_sptr init_translation = 
    new rgrl_trans_translation( trans, covariance_matrix );

  int k = 2;
  rgrl_matcher_sptr cp_matcher = new rgrl_matcher_k_nearest( k );
  rgrl_scale_sptr dummy_scale = new rgrl_scale();
  rgrl_match_set_sptr pruned_match_set = 
    cp_matcher->compute_matches(*moving_feature_set, 
                                *fixed_feature_set, 
                                *init_translation, 
                                moving_image_region, 
                                *dummy_scale);

  // EndCodeSnippet

  vcl_cout<<"pruned match set = "<<pruned_match_set->from_size()<<vcl_endl;

  //scale estimators
  int max_set_size = 500;  //maximum expected number of features
  vcl_auto_ptr<rrel_objective> muset_obj( new rrel_muset_obj( max_set_size , false) );

  rgrl_scale_estimator_unwgted_sptr unwgted_scale_est;
  rgrl_scale_estimator_wgted_sptr wgted_scale_est;

  unwgted_scale_est = new rgrl_scale_est_closest( muset_obj );
  wgted_scale_est = new rgrl_scale_est_all_weights();

  // BeginLatex
  //
  // Random sampling is done by
  // \code{rgrl\_initializer\_ran\_sam}. The estimation is triggered
  // when the IRLS registration engine asks for an initial
  // transformation.
  //
  // EndLatex
  
  // BeginCodeSnippet

  int affine_dof = 6;                  
  int affine_numSampleForFit = 3; 
  rgrl_estimator_sptr affine_estimator = 
    new rgrl_est_affine(affine_dof, 
                        affine_numSampleForFit);
  rgrl_initializer_ran_sam* ran_sam = new rgrl_initializer_ran_sam();
  ran_sam->set_sampling_params();
  ran_sam->set_data(pruned_match_set, 
                    unwgted_scale_est, 
                    moving_image_region,
                    affine_estimator,
                    0 );
  rgrl_initializer_sptr initializer = ran_sam;

  // EndCodeSnippet

  // BeginLatex
  //
  // To allow the similarity weight to be part of the final match
  // weight, we have to be careful when defining
  // \code{rgrl\_weighter\_m\_est}. Make sure Both
  // \code{use\_signature\_error} and \code{signature\_precomputed}
  // are allowed.
  //
  // EndLatex
  
  // BeginCodeSnippet
  
  vcl_auto_ptr<rrel_m_est_obj>  m_est_obj( new rrel_tukey_obj(4) );
  bool use_signature_error = false;
  bool signature_precomputed = true;
  rgrl_weighter_sptr wgter = new rgrl_weighter_m_est(m_est_obj, 
                                                     use_signature_error, 
                                                     signature_precomputed);

  // EndCodeSnippet

  // BeginLatex
  //
  // The rest of the components for running the IRLS registration
  // engine (\code{rgrl\_registration.run()}) are the same as in
  // \verb+registration2.cxx+.
  //
  // EndLatex

  // Transformation estimator for quadratic
  //
  int quad_dof = 12;                    //parameter degree of freedom
  int quad_numSampleForFit = 12;        //minimum number of samples for a fit
  rgrl_estimator_sptr quad_estimator = 
    new rgrl_est_quadratic(quad_dof, 
                           quad_numSampleForFit);


  // Convergence test
  //
  double tolerance = 1.5;
  rgrl_convergence_tester_sptr conv_test = 
    new rgrl_convergence_on_weighted_error( tolerance );


  fb_callback = callback;

  rgrl_data_manager_sptr data = new rgrl_data_manager();
  data->add_data( moving_feature_set,   // data from moving image
                  fixed_feature_set,    // data from fixed image
                  cp_matcher,           // matcher for this data
                  wgter,                // weighter
                  unwgted_scale_est,    // unweighted scale estimator
                  wgted_scale_est );    // weighted scale estimator
  data->add_estimator( quad_estimator );

  rgrl_feature_based_registration reg( data, conv_test );

  reg.run( initializer );

  // Output Results
  //
  if ( reg.has_final_transformation() ) {
    vcl_cout<<"Final xform: "<<vcl_endl;
    rgrl_transformation_sptr trans = reg.final_transformation();
    rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
    vcl_cout<<"Initial xform: Q \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
      "t = "<<q_xform->t()<<vcl_endl;
    vcl_cout<<"Final alignment error = "<<reg.final_status()->error()<<vcl_endl;
  }

  // BeginLatex
  //
  // Let's execute this example with the feature files
  // \verb+rgrl/data/IMG0002.pgm.landmars.txt+ for the moving image,
  // and \verb+rgrl/data/IMG0004.pgm.landmarks.txt+ for the fixed
  // image. The mask image is \verb+examples/mask.png+.
  //
  // EndLatex
  
  return 0;
}
