// BeginLatex
//
// This example is the dbicp registration algorithm with prior view.
// The algorithm can be found in \cite{stewart:tmi-03}. 
//
// EndLatex

#include <vcl_fstream.h>
#include <vcl_iostream.h>
#include <vnl/vnl_vector_fixed.h>
#include <vnl/vnl_matrix.h>

#include <vil/vil_load.h>
#include <vil/vil_image_view.h>

#include <rrel/rrel_muset_obj.h>
#include <rrel/rrel_tukey_obj.h>
#include <rgrl/rgrl_view_based_registration.h>
#include <rgrl/rgrl_view.h>
#include <rgrl/rgrl_view_gen_dbicp.h>
#include <rgrl/rgrl_feature_set_location_masked.h>
#include <rgrl/rgrl_matcher_k_nearest.h>
#include <rgrl/rgrl_trans_quadratic.h>
#include <rgrl/rgrl_est_quadratic.h>
#include <rgrl/rgrl_trans_similarity.h>
#include <rgrl/rgrl_est_similarity2d.h>
#include <rgrl/rgrl_trans_reduced_quad.h>
#include <rgrl/rgrl_est_reduced_quad2d.h>

#include <rgrl/rgrl_feature_trace_pt.h>
#include <rgrl/rgrl_scale_est_all_weights.h>
#include <rgrl/rgrl_scale_est_closest.h>
#include <rgrl/rgrl_weighter_m_est.h>
#include <rgrl/rgrl_convergence_on_weighted_error.h>

#include <rgrl/rgrl_roi.h>
#include <rgrl/rgrl_cast.h>
#include <rgrl/rgrl_mask.h>
#include <rgrl/rgrl_converge_status.h>
#include <rgrl/rgrl_data_manager.h>

typedef vcl_vector< rgrl_feature_sptr >  feature_vector;
typedef vnl_vector_fixed<double,2>       vector_2d;

void 
read_feature_file( const char* filename,
                   feature_vector& trace_points )
{
  vcl_ifstream istr( filename );

  if( !istr ) {
    vcl_cerr<<"ERROR: Cannot open "<<filename<<vcl_endl;
    return;
  }

  vector_2d location;
  vector_2d direction;

  bool done = false;
  while ( !done && istr ) {
    if ( !(istr >> location[0] >> location[1] >> direction[0] >> direction[1]) )
      done = true;
    else trace_points.push_back( new rgrl_feature_trace_pt(location, direction) );
  }

  istr.close();
  vcl_cout<<"There are "<<trace_points.size()<<" features"<<vcl_endl;

}

extern void (*vb_callback)( rgrl_view_sptr view, rgrl_set_of<rgrl_match_set_sptr> const& match_sets );
void callback( rgrl_view_sptr view, rgrl_set_of<rgrl_match_set_sptr> const& match_sets )
{
  rgrl_transformation_sptr trans = view->xform_estimate();

  if ( trans->type() == RGRL_SIMILARITY_XFORM ) {
    rgrl_trans_similarity* sim_xform = rgrl_cast<rgrl_trans_similarity*>(trans);
    vcl_cout<<"Initial xform: A \n"<<sim_xform->A()<<"\nt = "<<sim_xform->t()<<vcl_endl;}
  else if ( trans->type() == RGRL_REDUCED_QUAD_XFORM ) {
    rgrl_trans_reduced_quad* rq_xform = rgrl_cast<rgrl_trans_reduced_quad*>(trans);
    vcl_cout<<"Initial xform: Q \n"<<rq_xform->Q()<<"A = "<<rq_xform->A()<<
      "t = "<<rq_xform->t()<<vcl_endl;
  }
  else if ( trans->type() == RGRL_QUADRATIC_XFORM ) {
    rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
    vcl_cout<<"Initial xform: Q \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
      "t = "<<q_xform->t()<<vcl_endl;
  }
  else vcl_cout<<"Unkown type"<<vcl_endl;
 
 }

int 
main( int argc, char* argv[] )
{
  if( argc < 4 ) {
    vcl_cerr << "Missing Parameters " << vcl_endl;
    vcl_cerr << "Usage: " << argv[0];
    vcl_cerr << " FixedImageLandmarkFile (IMG0004) MovingImageLandmarkFile (IMG0002) MaskImage";
    return 1;
  }

  // Prepare feature sets
  //
  feature_vector moving_set;
  feature_vector fixed_set;
  const char* fixed_file_name = argv[1];
  const char* moving_file_name = argv[2];

  read_feature_file( moving_file_name, moving_set );
  read_feature_file( fixed_file_name, fixed_set );

  const unsigned int dimension = 2;
  rgrl_feature_set_sptr moving_feature_set;
  rgrl_feature_set_sptr fixed_feature_set;

  const char* make_file_name = argv[3];
  vil_image_view<vxl_byte> mask_image;
  mask_image = vil_load(make_file_name);
  rgrl_mask_sptr mask = new rgrl_mask_2d_image( mask_image );

  moving_feature_set = 
    new rgrl_feature_set_location_masked<dimension>(moving_set,
                                                    mask);
  fixed_feature_set = 
    new rgrl_feature_set_location_masked<dimension>(fixed_set,
                                                    mask);

  // Set the prior view
  //
  vector_2d x0(0,0);           //upper left corner
  vector_2d x1(1023,1023);     //bottom right corner
  rgrl_roi moving_image_region(x0, x1);
  rgrl_roi fixed_image_region(x0, x1);

  // Global region [ (0, 0) - (770.436, 1022.58) ]
  x0[0] = 289.202; x0[1] = 47.9876; 
  x1[0] = 1023; x1[1] = 1023;
  rgrl_roi global_image_region(x0, x1);
  // Initial region = ( 569.181 297.855 - 635.306 363.98 )
  x0[0] = 569.181; x0[1] = 297.855;
  x1[0] = 635.306; x1[1] = 363.98;
  rgrl_roi initial_image_region(x0, x1);

  /* IMG0004->IMg0002
  // Global region [ (0, 0) - (770.436, 1022.58) ]
  x1[0] = 770.436; x1[1] = 1022.58;
  rgrl_roi global_image_region(x0, x1);
  // Initial region [ (294.754, 261.995) - (357.754, 324.995) ]
  x0[0] = 294.754; x0[1] = 261.995;
  x1[0] = 357.754; x1[1] = 324.995;
  rgrl_roi initial_image_region(x0, x1);
  */

  // Initial transformation and estimator

  double shift_x = -302.301;
  double shift_y = -55.1425;
  vector_2d trans( shift_x, shift_y); 
  vnl_matrix<double> A(2,2);
  A(0,0) = 1.04599;       A(0,1) = -0.00390165;
  A(1,0) = 0.00390165;    A(1,1) =1.04599;

  /* IMG0004->IMG0002
  double shift_x = 290.064;
  double shift_y = 52.7184;
  vector_2d trans( shift_x, shift_y); 
  vnl_matrix<double> A(2,2);
  A(0,0) = 0.9529;         A(0,1) = 0.00408918;
  A(1,0) = -0.00408918;    A(1,1) = 0.9529;
  */

  vnl_matrix<double> covariance_matrix( 4, 4, vnl_matrix_identity );
  rgrl_transformation_sptr init_similarity = new rgrl_trans_similarity(A, trans, covariance_matrix);
  rgrl_estimator_sptr estimator_sim = new rgrl_est_similarity2d(4, 4);

  // Now the view
  rgrl_view_sptr prior_view = new rgrl_view( moving_image_region, fixed_image_region,
                                             initial_image_region, global_image_region,
                                             estimator_sim, init_similarity, 0);

  // Create the view generator, which is dbicp
  //
  rgrl_view_generator_sptr view_gen = new rgrl_view_gen_dbicp(2, 1.04, 1);
  

  // Creat the matcher for closest point matching.
  //
  rgrl_matcher_sptr cp_matcher = new rgrl_matcher_k_nearest( 1 );
  
  // robust weighting
  //
  vcl_auto_ptr<rrel_m_est_obj>  m_est_obj( new rrel_tukey_obj(4) );
  rgrl_weighter_sptr wgter = new rgrl_weighter_m_est( m_est_obj, false, false );
  
  // estimate scales based on closest match. Since we will only have
  // one-to-one matches (because of the choice in matcher above), the
  // closest match is _the_ match.
  //
  rgrl_scale_estimator_unwgted_sptr unwgted_scale_est;
  vcl_auto_ptr<rrel_objective> muset_obj( new rrel_muset_obj( 3000 , false) );
  rgrl_scale_est_closest* est = new rgrl_scale_est_closest( muset_obj );
  unwgted_scale_est = est;
  
  // how to determine when we are done
  //
  double tolerance = 1.5;
  rgrl_convergence_tester_sptr conv_test = 
    new rgrl_convergence_on_weighted_error( tolerance );
  
  // Add the data and model hierarchy needed by the view generator.
  //
  rgrl_data_manager_sptr data = new rgrl_data_manager();
  
  data->add_data( moving_feature_set,// from data
                  fixed_feature_set, // to data
                  cp_matcher,        // how to create correspondences for this data
                  wgter,             // weighter
                  unwgted_scale_est, // unweighted scale estimator
                  0);                // null weighted scale estimator

  data->add_estimator( estimator_sim );
  data->add_estimator( new rgrl_est_reduced_quad2d(6, 6) );
  data->add_estimator( new rgrl_est_quadratic(12, 12) );

  // Set up the reg algorithm
  //
  rgrl_view_based_registration reg( data, view_gen, conv_test );
  reg.set_max_icp_iter(5);
  reg.set_expected_max_geometric_scale(30);
  vb_callback = callback;

  reg.run( prior_view );

  // Output Results
  //
  if ( reg.has_best_view() ) {
    vcl_cout<<"Final xform: "<<vcl_endl;
    rgrl_transformation_sptr trans = reg.final_transformation();
    rgrl_trans_quadratic* q_xform = rgrl_cast<rgrl_trans_quadratic*>(trans);
    vcl_cout<<"Initial xform: Q \n"<<q_xform->Q()<<"A = "<<q_xform->A()<<
      "t = "<<q_xform->t()<<vcl_endl;
    vcl_cout<<"Final alignment error = "<<reg.final_status()->error()<<vcl_endl;
  }

  // BeginLatex
  //
  // Let's execute this example with the feature files
  // \verb+rgrl/data/IMG0002.pgm.txt+ for the moving image, and
  // \verb+rgrl/data/IMG0004.pgm.txt+ for the fixed image. The mask
  // image is \verb+examples/mask.png+.
  //
  // EndLatex

  return 0;
}
