#include "rgrl_convergence_on_median_error.h"

#include <vcl_vector.h>
#include <vcl_algorithm.h>
#include <vcl_cmath.h>

#include "rgrl_match_set.h"
#include "rgrl_set_of.h"
#include "rgrl_converge_status.h"
#include "rgrl_view.h"

#ifndef NDEBUG
#define DEBUG
#endif

#ifdef DEBUG
# include <vcl_iostream.h>
# define DBG(x) vcl_cout << x
#else
# define DBG(x) /* x */
#endif

rgrl_convergence_on_median_error::
rgrl_convergence_on_median_error( double tol )
  : tolerance_( tol )
{
}


rgrl_converge_status_sptr 
rgrl_convergence_on_median_error::
compute_status( rgrl_converge_status_sptr               prev_status,
                rgrl_view                        const& prev_view,
                rgrl_view                        const& current_view,
                rgrl_set_of<rgrl_match_set_sptr> const& current_match_sets,
                rgrl_set_of<rgrl_scale_sptr>     const& /*current_scales*/) const
{
  typedef rgrl_match_set::const_from_iterator from_iter;
  typedef from_iter::to_iterator              to_iter;

  // Step1: Compute the errors of all matches
  //
  rgrl_transformation_sptr current_xform = current_view.xform_estimate();
  vcl_vector<double> errors;

  for( unsigned ds=0; ds < current_match_sets.size(); ++ds ) {
    rgrl_match_set const& ms = *current_match_sets[ds];
    for( from_iter fitr = ms.from_begin(); fitr != ms.from_end(); ++fitr ) {
      rgrl_feature_sptr mapped = fitr.from_feature()->transform( *current_xform );
      for( to_iter titr = fitr.begin(); titr != fitr.end(); ++titr ) {
        errors.push_back( titr.to_feature()->geometric_error( *mapped ) );
      }
    }
  }

  // Step2: Take the median of the sorted error vector as the objective value
  //
  assert( errors.size() > 0 );
  vcl_vector<double>::iterator middle = errors.begin() + errors.size()/2;
  vcl_nth_element( errors.begin(), middle, errors.end() );

  double new_error = *middle;
  bool good = new_error < tolerance_;

  // Step3: First check if the regions in the view have converged. If yes,
  //        check the convergence of the transform estimate by comparing to 
  //        the previous status. The oscillation count is incremented if the 
  //        error_diff changes the sign.
  //
  bool converged;
  bool stagnated = false;
  unsigned int oscillation_count = 0;
  double error_diff = 0.0;
  if( prev_status && current_view.regions_converged_to(prev_view) ) {
    double old_error = prev_status->error(); 
    error_diff = new_error-old_error;
    converged = vcl_abs( (new_error-old_error) / new_error ) < 1e-4;

    if( !converged ) {
      // look for oscillation
      if( error_diff * prev_status->error_diff() < 0.0 ) {
        oscillation_count = prev_status->oscillation_count() + 1;
        DBG( "Oscillation. Count="<<oscillation_count<<"\n" );
      } else {
        if( prev_status->oscillation_count() > 0 )
          oscillation_count = prev_status->oscillation_count() - 1;
      }
      if( oscillation_count > 10 ) {
        stagnated = true;
      }
    }
  } else {
    converged = false;
  }

  return new rgrl_converge_status( converged, stagnated, good, new_error, oscillation_count, error_diff );
}
