#include "rgrl_feature_based_registration.h"

#include "rgrl_initializer.h"
#include "rgrl_convergence_tester.h"
#include "rgrl_converge_status.h"
#include "rgrl_estimator.h"
#include "rgrl_scale_estimator.h"
#include "rgrl_weighter_unit.h"
#include "rgrl_matcher.h"
#include "rgrl_match_set.h"
#include "rgrl_scale_est_null.h"
#include "rgrl_cast.h"
#include "rgrl_data_manager.h"
#include "rgrl_scale.h"
#include "rgrl_util.h"

const static unsigned verbose_ = 5;

#define TRACE( n, X ) \
  if( verbose_ > n ) vcl_cout << X
#include <vcl_iostream.h>

void (*fb_callback)(rgrl_transformation_sptr trans, rgrl_set_of<rgrl_match_set_sptr> const& matches) = 0;


rgrl_feature_based_registration::
rgrl_feature_based_registration( rgrl_data_manager_sptr data,
                                 rgrl_convergence_tester_sptr conv_tester )
  :data_( data ),
   conv_tester_( conv_tester ),
   num_xforms_tested_( 0 ),
   max_icp_iter_(25),
   expected_max_geometric_scale_ (0),
   expected_min_geometric_scale_ (0),
   iterations_for_scale_est_(-1)
{
}

rgrl_feature_based_registration::
~rgrl_feature_based_registration()
{
}

//: Running from multiple initial estimates, produced by the
//  initializer during registration Loop through the set of initial
//  estimates, and call the next \a run(.) in the loop
void 
rgrl_feature_based_registration::
run( rgrl_initializer_sptr initializer )
{
  //  Clear previous results
  num_xforms_tested_ = 0;
  best_xform_estimate_ = 0;
  best_matches_.clear();   
  best_scales_.clear();
  best_status_ = 0;

  rgrl_transformation_sptr          init_xform_estimate;
  rgrl_scale_sptr                   prior_scale;
  rgrl_estimator_sptr               init_xform_estimator;
  rgrl_roi                          image_region(0);
  unsigned                          init_resolution;

  while( initializer->next_initial( image_region, init_xform_estimator, 
                                    init_xform_estimate, init_resolution, 
                                    prior_scale ) ) {
    this->run( image_region, init_xform_estimator, init_xform_estimate, 
               prior_scale, init_resolution );

    if ( best_status_ && best_status_->is_good_enough() ) break;

    ++num_xforms_tested_;
  }
}

//: Running from a given initial estimate. 
//
//  Based on if data_->is_multi_feature(), call run_single_feature(.)
//  or run_multi_feature(.)
//
void 
rgrl_feature_based_registration::
run( rgrl_roi                   image_region,
     rgrl_estimator_sptr        init_xform_estimator,
     rgrl_transformation_sptr   initial_xform,
     rgrl_scale_sptr            prior_scale,  
     unsigned                   init_resolution)
{
  if ( data_->is_multi_feature() ) {
    TRACE( 3, " Multi-feature Registration: \n" );
    this->register_multi_feature(image_region, init_xform_estimator, 
                                 initial_xform, prior_scale, init_resolution);
  }
  else {
    TRACE( 3, " Single-feature Registration: \n" );
    this->register_single_feature(image_region, init_xform_estimator, 
                                  initial_xform, prior_scale, init_resolution);
  }
}

//////////////// functions to access internal data  ////////////////////////

//:  Return the final, best estimate
rgrl_transformation_sptr 
rgrl_feature_based_registration::
final_transformation() const
{
  return best_xform_estimate_;
}

//:  Return the scales of the best transformation estimate
rgrl_set_of<rgrl_scale_sptr> const& 
rgrl_feature_based_registration::
final_scales() const
{
  return best_scales_;
}

rgrl_scale_sptr 
rgrl_feature_based_registration::
final_scale() const
{
  assert( !data_->is_multi_feature() );
  return best_scales_[0];
}
  
//:  Return the status of the best transformation estimate.
rgrl_converge_status_sptr 
rgrl_feature_based_registration::
final_status() const
{
  return best_status_;
}

//:  The matches used for the best transformation estimate.
rgrl_set_of<rgrl_match_set_sptr>  const& 
rgrl_feature_based_registration::
final_match_sets() const
{
  return best_matches_;
}

rgrl_match_set_sptr 
rgrl_feature_based_registration::
final_match_set() const
{
  assert( !data_->is_multi_feature() );
  return best_matches_[0];
}

//:  Return the number of initial transformations tested
unsigned 
rgrl_feature_based_registration::
num_initial_xforms_tested() const
{
  return num_xforms_tested_;
}

//:  Return true is has a best xform_estimate
bool 
rgrl_feature_based_registration::
has_final_transformation() const
{
  if ( best_xform_estimate_ ) return true;
  return false;
}

//: Set the max number of icp iteration per level
void 
rgrl_feature_based_registration::
set_max_icp_iter( unsigned iter )
{
   max_icp_iter_ = iter;
}
  
//: Set the expected max scale
void 
rgrl_feature_based_registration::
set_expected_max_geometric_scale( double scale)
{
  expected_max_geometric_scale_ = scale;
}

//: Set the expected min scale
void 
rgrl_feature_based_registration::
set_expected_min_geometric_scale( double scale)
{
  expected_min_geometric_scale_ = scale;
} 

//: Set the number of iteration for scale estimation
void 
rgrl_feature_based_registration::
set_iterations_for_scale_est( int iter)
{
  iterations_for_scale_est_ = iter;
}

//////////////////// private functions //////////////////////////

//: registration of single feature type at each stage/resolution
void 
rgrl_feature_based_registration::
register_single_feature( rgrl_roi                 image_region,
                         rgrl_estimator_sptr      initial_xform_estimator,
                         rgrl_transformation_sptr xform_estimate,
                         rgrl_scale_sptr          scale,
                         unsigned                 resolution )
{  
  rgrl_converge_status_sptr         current_status;
  rgrl_feature_set_sptr             from_set;
  rgrl_feature_set_sptr             to_set;
  rgrl_matcher_sptr                 matcher;
  rgrl_scale_estimator_unwgted_sptr unwgted_scale_est;
  rgrl_scale_estimator_wgted_sptr   wgted_scale_est;
  rgrl_weighter_sptr                weighter;
  rgrl_match_set_sptr               match_set;
  vcl_vector<rgrl_estimator_sptr>   xform_estimators;
  rgrl_estimator_sptr               xform_estimator;
  bool                              failed, scale_in_range;
  unsigned                          prev_resol;

  failed = false;
  scale_in_range = true;

  do { // for each stage/resolution
    data_->get_data_at_stage( resolution, from_set, to_set, matcher, weighter, 
                              unwgted_scale_est, wgted_scale_est, xform_estimators);
    match_set = 0;
    
    TRACE( 1, " Current resolution "<< resolution <<"\n" );
    
    
    // If no estimator found for the current stage, the default is
    // from the initializer. Feature_based can only deal with one
    // estimator in each stage/resolution. If more than one is found,
    // only the first is kept.
    //
    if ( xform_estimators.empty() )
      xform_estimator = initial_xform_estimator;
    else xform_estimator = xform_estimators[0]; 
    
    // Check if the weighter is null. If so,set it to the unit
    // weighter, i.e. the weight is set to 1 for every match.
    //
    if ( !weighter )
      weighter = new rgrl_weighter_unit();
    
    // Check if all scale estimators are null. In this case, assign a
    // null scale estimator to the unwgted_scale_ests, i.e., the sale
    // is 1.
    //
    if ( !unwgted_scale_est && !wgted_scale_est )
      unwgted_scale_est = new rgrl_scale_est_null();
    
    unsigned iterations = 0; //keeps track of total iter at stage
    current_status = 0;
    bool should_estimate_scale = true;
    int  scale_est_count = 0;

    do { // for each re-match
      
      TRACE( 3, " Computing matches and scales\n" );
      // Compute matches, and scales for each feature set.
      //
      match_set= matcher->compute_matches( *from_set,
                                           *to_set,
                                           *xform_estimate,
                                           image_region,
                                           *scale );
      TRACE( 3, "      Matches: " << match_set->from_size() <<"\n" );
      
      // For the first iteration, use prior scale or estimate the
      // scales without weights (since we don't have any...)
      //
      if ( iterations_for_scale_est_ >= 0 && 
           scale_est_count > iterations_for_scale_est_ ) {
        should_estimate_scale = false;
      }

      rgrl_scale_sptr  new_scale = 0;
      if ( !should_estimate_scale ) {
        TRACE( 3, "No scale estimation \n" );   
      }
      else if ( match_set->from_size() == 0 ) {
        vcl_cerr<<"rgrl_feature_based_registration:: Empty match set!!!"<<vcl_endl;
        failed = true;
        continue;
      }
      else { //should_estimate_scale && match_set->from_size() > 0
        if( !wgted_scale_est ) {              
          new_scale = unwgted_scale_est->estimate_unweighted( *match_set, scale );
        } else {
          if( iterations == 0 && !scale ) {
            assert( unwgted_scale_est );
            new_scale = unwgted_scale_est->estimate_unweighted( *match_set, scale);
          }
          else {
            weighter->compute_weights(*scale, *match_set);
            new_scale = wgted_scale_est->estimate_weighted( *match_set, scale);
          }
        }

        // If the new scale exists, and the geometric scale is above
        // the lower bound, replace the old one by the new one
        if ( new_scale ) {
          TRACE( 3, "New geometric scale = "<<new_scale->geometric_scale()<<"\n" );
          if ( new_scale->has_geometric_scale() && 
               new_scale->geometric_scale() < expected_min_geometric_scale_ ) 
            should_estimate_scale = false;
          else {
            scale = new_scale;
            scale_est_count++;
          }
        }
      }
      assert( scale );

      TRACE( 3, " Estimate the transformation\n" );
      
      // Transformation estimation using a simplified irls
      // (Iterative-Reweighted Least-Squares) routine.
      //
      // The match sets and the scales are fixed, but the associated
      // weights are updated (hence reweighted-least-squares)
      // throughout the estimation.
      //
      double alignment_error;
      if ( !rgrl_util_irls( match_set, scale, weighter,
                            *conv_tester_, xform_estimator, xform_estimate, 
                            alignment_error) ) {
        failed = true;
        continue; //no valid xform, so exit the loop
      }

      // For debugging ...
      //
      rgrl_set_of<rgrl_match_set_sptr> match_sets;
      match_sets.push_back( match_set );
      if( fb_callback ) fb_callback( xform_estimate, match_sets );
      
      // Update the weights and scale estimates based on the new transform
      //
      TRACE( 3, " Updating scale estimates and checking for validity\n" );
      
      match_set->remap_from_features( *xform_estimate );
      weighter->compute_weights( *scale, *match_set );

      if ( should_estimate_scale ) {
        if( !wgted_scale_est )             
          scale = unwgted_scale_est->estimate_unweighted( *match_set, scale );
        else 
          scale = wgted_scale_est->estimate_weighted( *match_set, scale );
        TRACE( 3, " Geometric scale after converged = "<< scale->geometric_scale()<<"\n" );
      }
      
      if ( expected_max_geometric_scale_ > 0 && 
           scale->has_geometric_scale() && 
           scale->geometric_scale() > expected_max_geometric_scale_) {
        scale_in_range = false;
        failed = true;
      }
      
      // Perform convergence test
      //
      TRACE( 3, " Perform convergence test\n" );
      current_status = 
        conv_tester_->compute_status( current_status,
                                      xform_estimate, 
                                      xform_estimator,
                                      match_set, scale );
      TRACE( 4, "run: (iterations = " << iterations << ") oscillation count = " << current_status->oscillation_count() << "\n" );
      TRACE( 4, "run: error = " << current_status->error() << vcl_endl );
      TRACE( 4, "run: error_diff = " << current_status->error_diff() << vcl_endl );
      
      ++iterations;
    } while( !failed &&
             !current_status->has_converged() &&
             !current_status->has_stagnated() &&
             iterations < max_icp_iter_ );
    
    if ( failed ) {
      if ( !scale_in_range )
        TRACE( 3, " Geometric scale above the expected value \n" );
      else TRACE( 3, " Failed with empty match set, or irls estimation\n" );
      continue;
    }
    if ( current_status->has_converged() )
      TRACE( 3, " CONVERGED \n" );
    if ( current_status->has_stagnated() )
      TRACE( 3, " STAGNATED \n" );
    if ( iterations == max_icp_iter_ ) 
      TRACE( 3, " ICP iteration reached maximum ("<<max_icp_iter_<<" )\n" );
    
    // Move to the next resolution with proper initialization if not
    // at the finest level
    //
    prev_resol = resolution;
    initialize_for_next_resolution( image_region, xform_estimate, 
                                    resolution );
    
  } while ( !failed && 
            !resolution == 0 ||  
            !prev_resol == 0 );
  
  TRACE( 3, "Estimation complete\n" );
  
  //   At this point the iterations are over for this initial
  //   estimate.  If the estimate is successful, if it is the first
  //   one tested or it is the best seen thus far, record information
  //   about this estimate.
  //
  
  if ( !failed ){
    if ( !best_status_ ||
         current_status->objective_value() < best_status_->objective_value() ) {
      best_xform_estimate_ = xform_estimate;
      if (best_scales_.size() != 1)
        best_scales_.resize(1);
      best_scales_[0] = scale;
      if (best_matches_.size() != 1)
        best_matches_.resize(1);
      best_matches_[0] = match_set;
      best_status_ = current_status;
      TRACE( 3, "Set best xform estimate\n" );
    }
  }
}

//: registration of multiple feature type at each stage/resolution
void
rgrl_feature_based_registration::
register_multi_feature( rgrl_roi                 image_region,
                        rgrl_estimator_sptr      initial_xform_estimator,
                        rgrl_transformation_sptr xform_estimate,
                        rgrl_scale_sptr          prior_scale,
                        unsigned                 resolution )
{
  rgrl_converge_status_sptr                     current_status;
  vcl_vector<rgrl_feature_set_sptr>             from_sets;
  vcl_vector<rgrl_feature_set_sptr>             to_sets;
  vcl_vector<rgrl_matcher_sptr>                 matchers;
  vcl_vector<rgrl_scale_estimator_unwgted_sptr> unwgted_scale_ests;
  vcl_vector<rgrl_scale_estimator_wgted_sptr>   wgted_scale_ests;
  vcl_vector<rgrl_weighter_sptr>                weighters;
  rgrl_set_of<rgrl_match_set_sptr>              match_sets;
  rgrl_set_of<rgrl_scale_sptr>                  scales;
  vcl_vector<rgrl_estimator_sptr>               xform_estimators;
  rgrl_estimator_sptr                           xform_estimator;
  bool                                          failed, scale_in_range, use_prior_scale;
  unsigned                                      prev_resol;

  use_prior_scale = false;
  if ( prior_scale ) use_prior_scale = true;
  scale_in_range = true;
  failed = false;

  do { // for each stage/resolution
    data_->get_data_at_stage( resolution, from_sets, to_sets, matchers, 
                              weighters, unwgted_scale_ests, 
                              wgted_scale_ests, xform_estimators);
    unsigned data_count = from_sets.size();

    TRACE( 1, " Current resolution "<< resolution <<"\n" );

    // Initialize the size of match_sets and scales to be the same as
    // the from_sets
    //
    match_sets.clear();
    scales.clear();
    match_sets.resize( data_count );
    scales.resize( data_count );
    
    // If no estimator found for the current stage, the default is
    // from the initializer. Feature_based can only deal with one
    // estimator in each stage/resolution. If more than one is found,
    // only the first is kept.
    //
    if ( xform_estimators.size() == 0 )
      xform_estimator = initial_xform_estimator;
    else xform_estimator = xform_estimators[0]; 

    // Check if the weighter and scale estimators are set. If the
    // weighter is null, set it to the unit weighter, i.e. the weight
    // is set to 1 for every match. If no scale estimators are given,
    // set the unwgted one to null, i.e., the sale is 1.
    //
    for( unsigned int fs=0; fs < data_count; ++fs ) {
      if ( !weighters[fs] )
        weighters[fs] = new rgrl_weighter_unit();
      if ( !unwgted_scale_ests[fs] && !wgted_scale_ests[fs] )
        unwgted_scale_ests[fs] = new rgrl_scale_est_null();
    }

    // If the initialization comes with a prior scale, initialize the
    // scales using the prior scale.
    //
    if ( use_prior_scale ) {
      TRACE( 3, "Prior scale = "<<prior_scale->geometric_scale()<<"\n" );
      for (  unsigned int fs = 0; fs < data_count; ++fs )
        scales[fs] = prior_scale;
    }

    unsigned iterations = 0; //keeps track of total iter at level
    current_status = 0;
    bool should_estimate_scale = true;
    int  scale_est_count = 0;

    do { // for each re-matching
      
      TRACE( 3, " Computing matches and scales\n" );
      // Compute matches, and scales for each feature set.
      //
      for( unsigned int fs=0; fs < data_count; ++fs ) {
        TRACE( 3, "   Data set " << fs << vcl_endl );
        rgrl_match_set_sptr new_matches =
          matchers[fs]->compute_matches( *from_sets[fs],
                                         *to_sets[fs],
                                         *xform_estimate,
                                         image_region,
                                         *scales[fs] );
        TRACE( 3, "      Matches: " << new_matches->from_size() <<" ("<<new_matches<<")\n" );
        
        // For the first iteration, use prior scale or estimate the
        // scales without weights (since we don't have any...)
        //
        if ( iterations_for_scale_est_ >= 0 && 
             scale_est_count > iterations_for_scale_est_ ) {
          should_estimate_scale = false;
        }

        rgrl_scale_sptr new_scale = 0;
        if ( !should_estimate_scale ) {
          TRACE( 3, "No scale estimation \n" );
        }
        else if ( new_matches->from_size() == 0 ) {
          vcl_cerr<<"rgrl_feature_based_registration:: Empty match set!!!"<<vcl_endl;
          failed = true;
          continue;
        }
        else { //should_estimate_scale && new_matches->from_size() > 0 
          if( !wgted_scale_ests[fs] ) {              
            new_scale = unwgted_scale_ests[fs]->estimate_unweighted( *new_matches, scales[fs] );
          } else {
            if( iterations == 0 && !scales[fs] ) {
              assert( unwgted_scale_ests[fs] );
              new_scale = unwgted_scale_ests[fs]->estimate_unweighted( *new_matches, scales[fs]);
            }
            else {
              weighters[fs]->compute_weights(*scales[fs], *new_matches);
              new_scale = wgted_scale_ests[fs]->estimate_weighted( *new_matches, scales[fs]);
            }
          }
          
          if ( new_scale ) {
            TRACE( 3, "New geometric scale = "<<new_scale->geometric_scale()<<"\n" );
            if ( new_scale->has_geometric_scale() && 
                 new_scale->geometric_scale() < expected_min_geometric_scale_ ) 
              should_estimate_scale = false;
            else {
              scale_est_count++;
              scales[fs] = new_scale;
            }
          }
        }
        assert( scales[fs] );

        // Keep new ones and discard old ones
        //
        match_sets[fs] = new_matches;
      }

      TRACE( 3, " Estimate the transformation\n" );
      
      // Estimate the transformation using a simplified irls
      // (Iterative-Reweighted Least-Squares) routine.
      //
      // The match sets and the scales are fixed, but the associated
      // weights are updated (hence reweighted-least-squares)
      // throughout the estimation.
      double alignment_error;
      if ( !rgrl_util_irls( match_sets, scales, weighters,
                            *conv_tester_, xform_estimator, xform_estimate, 
                            alignment_error) ) {
        failed = true;
        continue; //no valid xform, so exit the loop
      }

      // For debugging ...
      //
      if( fb_callback ) fb_callback( xform_estimate, match_sets );

      // Update the weights and scale estimates based on the new transform
      //
      TRACE( 3, " Updating scale estimates and checking for validity\n" );
      for( unsigned int fs=0; fs < data_count; ++fs ) {
        if ( match_sets[fs]->from_size() > 0 ) {
          match_sets[fs]->remap_from_features( *xform_estimate );
          weighters[fs]->compute_weights( *scales[fs], *match_sets[fs] );
          
          if ( should_estimate_scale ) {
            if( !wgted_scale_ests[fs] )             
              scales[fs] = unwgted_scale_ests[fs]->estimate_unweighted( *match_sets[fs], scales[fs] );
            else scales[fs] = wgted_scale_ests[fs]->estimate_weighted( *match_sets[fs], scales[fs] );
            TRACE( 3, " Geometric scale after converged = "<< scales[fs]->geometric_scale()<<"\n" );
          }
          
          if ( expected_max_geometric_scale_ > 0 && 
               scales[fs]->has_geometric_scale() && 
               scales[fs]->geometric_scale() > expected_max_geometric_scale_) {
            scale_in_range = false;
            failed = true;
          }
        }
      }

      // Perform convergence test
      //
      TRACE( 3, " Perform convergence test\n" );
      current_status = 
        conv_tester_->compute_status( current_status,
                                      xform_estimate, 
                                      xform_estimator,
                                      match_sets, scales );
      TRACE( 4, "run: (iterations = " << iterations << ") oscillation count = " << current_status->oscillation_count() << "\n" );
      TRACE( 4, "run: error = " << current_status->error() << vcl_endl );
      TRACE( 4, "run: error_diff = " << current_status->error_diff() << vcl_endl );
      
      ++iterations;
    } while( !failed &&
             !current_status->has_converged() &&
             !current_status->has_stagnated() &&
             iterations < max_icp_iter_ );
    
      if ( failed ) {
        if ( !scale_in_range )
          TRACE( 3, " Geometric scale above the expected value \n" );
        else TRACE( 3, " Failed with empty match set, or feature_based\n" );
        continue;
      }
      if ( current_status->has_converged() )
        TRACE( 3, " CONVERGED \n" );
      if ( current_status->has_stagnated() )
        TRACE( 3, " STAGNATED \n" );
      if ( iterations == max_icp_iter_ ) 
        TRACE( 3, " ICP iteration reached maximum ("<<max_icp_iter_<<" )\n" );
      
      // Move to the next resolution with proper initialization if not
      // already at the finest level
      //
      prev_resol = resolution;
      initialize_for_next_resolution( image_region, xform_estimate, 
                                      resolution );
      use_prior_scale = true;
      prior_scale = scales[0]; //Assuming scales[0] is a good approximate

  } while ( !failed && 
            !resolution == 0 ||  
            !prev_resol == 0 );
  
  TRACE( 3, "Estimation complete\n" );
  
  //   At this point the iterations are over for this initial
  //   estimate.  If the estimation is successful, and if it is the
  //   first one tested or it is the best seen thus far, record
  //   information about this estimate.
  //
  
  if ( !failed ){
    if ( !best_status_ ||
         current_status->objective_value() < best_status_->objective_value() ) {
      best_xform_estimate_ = xform_estimate;
      best_scales_         = scales;
      best_matches_        = match_sets;
      best_status_         = current_status;
      TRACE( 3, "Set best xform estimate\n" );
    }
  }
}


void
rgrl_feature_based_registration::
initialize_for_next_resolution(  rgrl_roi                 & image_region,
                                 rgrl_transformation_sptr & xform_estimate,
                                 unsigned                 & current_resol ) const
{
  // Find the next available level
  //
  unsigned new_resol = current_resol;
  int next_resol = new_resol - 1;
  for ( ; next_resol >= 0; --next_resol) {
    if ( data_->has_stage(next_resol) ) { 
      new_resol = next_resol;
      break;
    }
  }

  // If no next available level return
  //
  if ( new_resol == current_resol ) return;
 

  // Scale the components of the current view for the initial view of the next
  // available level
  //
  int level_diff = current_resol - new_resol;
  double dim_increase = data_->dimension_increase_for_next_stage(current_resol);
  double scale = vcl_pow(dim_increase, level_diff);

  image_region.set_x0( image_region.x0()*scale );
  image_region.set_x1( image_region.x1()*scale );

  xform_estimate =  xform_estimate->scale_by( scale );
  current_resol = new_resol;
}
