#ifndef rgrl_mask_h_
#define rgrl_mask_h_

//:
// \file
// \brief  hierarchy of mask of various shapes. 
// \author Charlene Tsai
// \date   Oct 2003

#include <vbl/vbl_ref_count.h>

#include <vil/vil_image_view.h>
#include <vnl/vnl_vector.h>
#include <vnl/vnl_math.h>

#include "rgrl_mask_sptr.h"
#include <vil3d/vil3d_image_view.h>

// Defines a ROI (termed mask) which is required by certain feature_set
// to outline region where operations are performed (e.g. registration).

//: base class of roi which is a pure virtual class
class rgrl_mask: public vbl_ref_count{
public:
  rgrl_mask() {}
  virtual 
  bool inside( vnl_vector<double> const& ) const = 0;
};


//: this is a wrapper on the binary mask image. The image is define for the x-y plane.
class rgrl_mask_2d_image: public rgrl_mask {
public:
  rgrl_mask_2d_image( const vil_image_view<vxl_byte>& in_mask, 
                      int org_x = 0, int org_y = 0)
    : mask_image_(in_mask), org_x_(org_x), org_y_(org_y) {}
  virtual bool inside( vnl_vector<double> const& pt ) const
  {
    int x = int(pt[0]-org_x_);
    int y = int(pt[1]-org_y_);
    bool in_range = ( 0<=vcl_floor((double)x) ) && ( 0<=vcl_floor((double)y) ) 
      && ( vcl_ceil((double)x)<mask_image_.ni() ) && ( vcl_ceil((double)y)<mask_image_.nj() );

    return in_range && mask_image_( x, y )>0;
  }
private:
  vil_image_view<vxl_byte> mask_image_;
  int org_x_, org_y_;
};

class rgrl_mask_3d_image: public rgrl_mask {
public:
  rgrl_mask_3d_image( const vil3d_image_view< vxl_byte > & in_mask,
                      int org_x = 0, int org_y = 0, int org_z = 0 )
    : mask_image_( in_mask ), org_x_( org_x ), org_y_( org_y ), org_z_( org_z ) {}
  virtual bool inside( vnl_vector< double > const& pt ) const
    {
      int x = int(pt[0]-org_x_);
      int y = int(pt[1]-org_y_);
      int z = int(pt[2]-org_z_);
      
      return ( 0 <= x ) && ( 0 <= y ) && ( 0 <= z ) &&
        ( x < (int)mask_image_.ni() ) && ( y < (int)mask_image_.nj() ) && ( z < (int)mask_image_.nk() ) && 
        mask_image_( x, y, z ) > 0 ;
    }

private:
  vil3d_image_view< vxl_byte > mask_image_;
  int org_x_, org_y_, org_z_;

};

//: This is a wrapper on the circular region or sphere
class rgrl_mask_sphere: public rgrl_mask{
public:
  rgrl_mask_sphere( const vnl_vector<double>& in_center,
                    const double& in_radius )
    : center_(in_center), radius_sqr_(in_radius*in_radius) {}
  virtual bool inside( vnl_vector<double> const& pt ) const
  {
    double sqr_sum = 0;
    for (unsigned int i = 0; i<pt.size(); ++i) {
      sqr_sum +=  vnl_math_sqr( pt[i] - center_[i] );
    }
    return sqr_sum < radius_sqr_ ;
  }

private:
  vnl_vector<double> center_;
  double radius_sqr_;
};


//: this is a wrapper on the box
class rgrl_mask_box:public rgrl_mask {
public:
  rgrl_mask_box( vnl_vector<double> const& x0, vnl_vector<double> const& x1 )
    : x0_(x0), x1_(x1) {}
  virtual bool inside( vnl_vector<double> const& pt ) const
  {
    bool inside = true;
    for (unsigned i =0; i<pt.size(); ++i) {
      inside = inside && (x0_[i] <= pt[i] && pt[i] <= x1_[i]);
    }
    return inside;
  }
private:
  vnl_vector<double> x0_;
  vnl_vector<double> x1_;
};

#endif

