#ifndef RGRL_SCOPED_PTR_H_
#define RGRL_SCOPED_PTR_H_

//:
// \file
// \author Amitha Perera
// \date   Dec 2002

//: Scoped pointer.
//
// This is a smart pointer that will release the resource when the
// scope of the pointer ends. This can be used to hold pointer members
// in a class without having to worry about resource leaks,
// etc. vcl_auto_ptr could be used for the same purpose, except that
// the reset() method is broken with MSVC6, and so it is not portable
// enough.
//
// Using this typically implies that your destructor is trivial (you
// don't need to explicitly delete memory) and that your code doesn't
// leak in the face of exceptions. However, to decrease dependencies,
// you must declare the destructor. Don't let the compiler do it for
// you. (Illustration below.)
//
// \warning Do not use this for arrays! The destructor calls delete,
// not delete[].
//
// Unlike normal pointers, this smart pointer "pushes in" constness, so that
// \code
//   rgrl_scoped_ptr<X> const
// \endcode
// behaves like
// \code
//   X const *;   // equivalently, const X *;
// \endcode
// instead of
// \code
//   X* const;
// \endcode
//
// I think this is often the intuitively required behaviour. This
// avoids the need for rgrl_scoped_ptr<X const> and all the associated
// type conversion headaches.
//
// When using this for member variables, you should declare and define
// a destructor for your class to avoid dependencies. For example, compare
// \code
//   // File b.h:
//   class X;
//   struct B
//   {
//     B();
//     vbl_scoped_ptr<X> m;
//   };
//
//   // File b.cxx:
//   #include "X.h"
//   B::B() : m( new X ) { }
//   // code for other members of B
//
//   // File main.cxx:
//   #include "b.h"
//   {
//     B a;
//   }
//   // a is out of scope here, and must be destroyed. Since B doesn't
//   // declare a destructor, the compiler will just call the destructors
//   // of all the members. This requires that we see the destructor for
//   // m, which in turn requires that we see the destructor for X, which
//   // means that the forward declaration of X in b.h is not enough.
// \endcode
// with
// \code
//   // File b.h:
//   class X;
//   struct B
//   {
//     B();
//     ~B();
//     vbl_scoped_ptr<X> m;
//   };
//
//   // File b.cxx:
//   #include "X.h"
//   B::B() : m( new X ) { }
//   B::~B()
//   {
//   } // the compiler will destruct m here, and therefore must see X here.
//     // Which is okay.
//
//   // code for members of B
//
//   // File main.cxx:
//   #include "b.h"
//   {
//     B a;
//   }
//   // a is out of scope here, and must be destroyed. Call B's destructor.
//   // No need to see X. This file is therefore not dependent on X.h. Good.
// \endcode
//
template<class T>
class rgrl_scoped_ptr
{
public:
  //:
  rgrl_scoped_ptr()
       : ptr_( 0 )
    {
    }

  //:
  // The scoped pointer will take ownership of the data
  rgrl_scoped_ptr( T* obj_ptr )
       : ptr_( obj_ptr )
    {
    }

  //:
  // This will destroy the object that the scoped pointer is pointing
  // to.
  ~rgrl_scoped_ptr()
    {
      delete ptr_;
    }

  //: Reset the pointer to a different object
  //
  // This will destory the current object (if any) and take ownership
  // of the data pointed to by \a obj_ptr.
  //
  void reset( T* obj_ptr )
    {
      delete ptr_;
      ptr_ = obj_ptr;
    }

  //: A "normal" pointer to the object
  //
  // This does not relinquish ownership; the scoped pointer still owns
  // the object.
  //
  T* get()
    {
      return ptr_;
    }

  //: A "normal" pointer to the object
  //
  // This does not relinquish ownership; the scoped pointer still owns
  // the object.
  //
  T const* get() const
    {
      return ptr_;
    }

  //:
  T& operator*()
    {
      return *ptr_;
    }

  //:
  T const& operator*() const
    {
      return *ptr_;
    }

  //:
  T* operator->()
    {
      return ptr_;
    }

  //:
  T const* operator->() const
    {
      return ptr_;
    }

  //:
  operator bool() const
    {
      return ptr_ != (T*)0;
    }

  //: Swap ownership of the data
  void swap( rgrl_scoped_ptr& other )
    {
      T* tmp = ptr_;
      ptr_ = other.ptr_;
      other.ptr_ = tmp;
    }

private:
  T* ptr_;

  // copy and assignment not allowed. This is not an auto_ptr!
  rgrl_scoped_ptr( rgrl_scoped_ptr const& );
  rgrl_scoped_ptr operator=( rgrl_scoped_ptr const& );
};

#endif // RGRL_SCOPED_PTR_H_
