//:
// \file
// \author Amitha Perera
// \date   Feb 2003
//
// \verbatim
// Modifications
//     Sep 2003 Charlene Tsai, added inv_map(.) functions.
// \endverbatim 

#include "rgrl_trans_affine.h"


#include <vcl_cassert.h>
#include <vnl/algo/vnl_svd.h>

rgrl_trans_affine::
rgrl_trans_affine( unsigned int dimension )
  : A_( vnl_matrix<double>( dimension, dimension, vnl_matrix_identity ) ),
    trans_( vnl_vector<double>( dimension, 0.0 ) ),
    covar_( vnl_matrix<double>( dimension*(dimension+1), dimension*(dimension+1), 0.0 ) )
{
}
  

rgrl_trans_affine::
rgrl_trans_affine( vnl_matrix<double> const& in_A, 
                   vnl_vector<double> const& in_trans,
                   vnl_matrix<double> const& in_covar )
  : A_( in_A ),
    trans_( in_trans ),
    covar_( in_covar ),
    from_centre_( in_trans.size(), 0.0 )
{
  assert( A_.rows() == A_.cols() );
  assert( A_.rows() == trans_.size() );
  assert( covar_.rows() == covar_.cols() );
  assert( covar_.rows() == A_.rows()* (A_.rows()+1) );
}

rgrl_trans_affine::
rgrl_trans_affine( vnl_matrix<double> const& in_A, 
                   vnl_vector<double> const& in_trans )
  : A_( in_A ),
    trans_( in_trans ),
    from_centre_( in_trans.size(), 0.0 )
{
  int dim = in_trans.size();
  covar_ = vnl_matrix<double>( dim*(dim+1), dim*(dim+1), 0.0 );

  assert( A_.rows() == A_.cols() );
  assert( A_.rows() == trans_.size() );
  assert( covar_.rows() == covar_.cols() );
  assert( covar_.rows() == A_.rows()* (A_.rows()+1) );
}

rgrl_trans_affine::
rgrl_trans_affine( vnl_matrix<double> const& in_A, 
                   vnl_vector<double> const& in_trans,
                   vnl_matrix<double> const& in_covar,
                   vnl_vector<double> const& in_from_centre,
                   vnl_vector<double> const& in_to_centre )
  : A_( in_A ),
    trans_( in_trans + in_to_centre ),
    covar_( in_covar ),
    from_centre_( in_from_centre )
{
  assert( A_.rows() == A_.cols() );
  assert( A_.rows() == trans_.size() );
  assert( covar_.rows() == covar_.cols() );
  assert( covar_.rows() == A_.rows()* (A_.rows()+1) );
  assert( from_centre_.size() == trans_.size() );
}


void
rgrl_trans_affine::
map_loc( vnl_vector<double> const& from,
         vnl_vector<double>      & to   ) const
{
  assert( from.size() == A_.rows() );
  to = A_ * (from-from_centre_) + trans_;
}


void
rgrl_trans_affine::
map_dir( vnl_vector<double> const& from_loc,
         vnl_vector<double> const& from_dir,
         vnl_vector<double>      & to_dir    ) const
{
  assert( from_loc.size() == A_.cols() );
  assert( from_dir.size() == A_.cols() );
  to_dir = A_ * from_dir;
  to_dir.normalize();
}


vnl_matrix<double>
rgrl_trans_affine::
transfer_error_covar( vnl_vector<double> const& p  ) const
{
  unsigned const m = A_.rows();

  assert( p.size() == m );

  vnl_matrix<double> temp( m, m*(m+1), 0.0 );
  for( unsigned i=0; i < m; ++i ) {
    unsigned off = i*(m+1);
    for( unsigned j=0; j < m; ++j ) {
      temp(i,off+j) = p[j] - from_centre_[j];
    }
    temp(i,off+m) = 1.0;
  }

  return temp * covar_ * temp.transpose();
}


vnl_matrix<double>
rgrl_trans_affine::
covar() const
{
  return covar_;
}


rgrl_transformation_type
rgrl_trans_affine::
type() const
{
  return RGRL_AFFINE_XFORM;
}


vnl_matrix<double> const&
rgrl_trans_affine::
A() const
{
  return A_;
}


vnl_vector<double>
rgrl_trans_affine::
t() const
{
  return trans_ - A_ * from_centre_;
}


void
rgrl_trans_affine::
inv_map( const vnl_vector<double>& to, 
         bool initialize_next,
         const vnl_vector<double>& to_delta,
         vnl_vector<double>& from,
         vnl_vector<double>& from_next_est) const
{
  const double epsilon = 0.01;
  vnl_vector<double> to_est = this->map_location(from);
  
  // compute the inverse of the Jacobian, which is the A_^-1
  vnl_svd<double> svd( A_ );
  vnl_matrix<double> J_inv = svd.inverse();

  // update from to become true inv_map of to, based on A^-1 and (to - to_est)
  if (vnl_vector_ssd(to, to_est) > epsilon*epsilon) {
    from += J_inv * (to - to_est);
  }
  if ( initialize_next ) {
    from_next_est = from + (J_inv * to_delta);
  }
}

void
rgrl_trans_affine::
inv_map( const vnl_vector<double>& to, 
         vnl_vector<double>& from ) const
{
  vnl_svd<double> svd( A_ );
  from = svd.inverse()*to - svd.inverse()*trans_ + from_centre_;
}

rgrl_transformation_sptr 
rgrl_trans_affine::
scale_by( double scale ) const
{
  return new rgrl_trans_affine( A_, trans_ * scale, 
                                covar_, from_centre_ * scale,
                                vnl_vector<double>(from_centre_.size(), 0.0) );
}
