#include "rgrl_view_based_registration.h"

#include "rgrl_initializer.h"
#include "rgrl_view_generator.h"
#include "rgrl_view.h"
#include "rgrl_convergence_tester.h"
#include "rgrl_converge_status.h"
#include "rgrl_estimator.h"
#include "rgrl_scale_estimator.h"
#include "rgrl_weighter_unit.h"
#include "rgrl_matcher.h"
#include "rgrl_match_set.h"
#include "rgrl_scale_est_null.h"
#include "rgrl_util.h"
#include "rgrl_scale.h"
#include "rgrl_data_manager.h"

#include <vnl/vnl_math.h>
#include "rgrl_cast.h"

const static unsigned verbose_ = 5;

#define TRACE( n, X ) \
  if( verbose_ > n ) vcl_cout << X
#include <vcl_iostream.h>

void (*vb_callback)( rgrl_view_sptr view, rgrl_set_of<rgrl_match_set_sptr> const& match_sets ) = 0;


rgrl_view_based_registration::
rgrl_view_based_registration( rgrl_data_manager_sptr        data,
                              rgrl_view_generator_sptr      view_generator,
                              rgrl_convergence_tester_sptr  conv_tester )
  :data_( data ),
   view_generator_( view_generator ),
   conv_tester_( conv_tester ),
   num_xforms_tested_( 0 ),
   max_icp_iter_(25),
   expected_max_geometric_scale_ (0),
   expected_min_geometric_scale_ (0),
   iterations_for_scale_est_(-1)
{
}


rgrl_view_based_registration::
~rgrl_view_based_registration()
{
}

//: Running from multiple initial estimates, produced by the
//  initializer during registration Loop through the set of initial
//  estimates, and call the next \a run(.) in the loop
void
rgrl_view_based_registration::
run( rgrl_initializer_sptr initializer )
{
  //  Clear previous results
  num_xforms_tested_ = 0;
  best_view_ = 0;
  best_matches_.clear();   
  best_scales_.clear();
  best_status_ = 0;

  rgrl_view_sptr                    initial_view;
  rgrl_scale_sptr                   prior_scale;

  while( initializer->next_initial( initial_view, prior_scale ) ) {
    this->run( initial_view, prior_scale );

    if ( best_status_ && best_status_->is_good_enough() ) break;

    ++num_xforms_tested_;
  }
}

//: Running from a given initial estimate. 
//
//  Based on if data_->is_multi_feature(), call run_single_feature(.)
//  or run_multi_feature(.)
//
void
rgrl_view_based_registration::
run( rgrl_view_sptr            prior_view,
     rgrl_scale_sptr           prior_scale )
{
  if ( data_->is_multi_feature() ) {
    TRACE( 3, " Multi-feature Registration: \n" );
    this->register_multi_feature( prior_view, prior_scale );
  }
  else {
    TRACE( 3, " Single-feature Registration: \n" );
    this->register_single_feature( prior_view, prior_scale );
  }
}

//////////////// functions to access internal data  ////////////////////////

//:  Return the final view for the best transformation estimate
rgrl_view_sptr
rgrl_view_based_registration::
final_view() const
{ 
  return best_view_;
}

//:  Return true if has a best view
bool
rgrl_view_based_registration::
has_best_view() const
{   
  if ( best_view_ ) return true;
  return false;
}

//:  Return the final, best estimate
rgrl_transformation_sptr 
rgrl_view_based_registration::
final_transformation() const
{
  return best_view_->xform_estimate();
}

//:  Return the scales of the best transformation estimate
rgrl_set_of<rgrl_scale_sptr> const& 
rgrl_view_based_registration::
final_scales() const
{
  return best_scales_;
}

rgrl_scale_sptr 
rgrl_view_based_registration::
final_scale() const
{
  assert( !data_->is_multi_feature() );
  return best_scales_[0];
}
  
//:  Return the status of the best transformation estimate.
rgrl_converge_status_sptr 
rgrl_view_based_registration::
final_status() const
{
  return best_status_;
}

//:  The matches used for the best transformation estimate.
rgrl_set_of<rgrl_match_set_sptr>  const& 
rgrl_view_based_registration::
final_match_sets() const
{
  return best_matches_;
}

rgrl_match_set_sptr 
rgrl_view_based_registration::
final_match_set() const
{
  assert( !data_->is_multi_feature() );
  return best_matches_[0];
}

//:  Return the number of initial transformations tested
unsigned 
rgrl_view_based_registration::
num_initial_xforms_tested() const
{
  return num_xforms_tested_;
}

//: Set the max number of icp iteration per level
void 
rgrl_view_based_registration::
set_max_icp_iter( unsigned iter )
{
   max_icp_iter_ = iter;
}
  
//: Set the expected max scale
void 
rgrl_view_based_registration::
set_expected_max_geometric_scale( double scale)
{
  expected_max_geometric_scale_ = scale;
}

//: Set the expected min scale
void 
rgrl_view_based_registration::
set_expected_min_geometric_scale( double scale)
{
  expected_min_geometric_scale_ = scale;
} 

//: Set the number of iteration for scale estimation
void 
rgrl_view_based_registration::
set_iterations_for_scale_est( int iter)
{
  iterations_for_scale_est_ = iter;
}
  
//////////////////// private functions //////////////////////////

//: registration of single feature type at each stage/resolution
void 
rgrl_view_based_registration::
register_single_feature( rgrl_view_sptr           initial_view,
                         rgrl_scale_sptr          scale )
{
  rgrl_view_sptr                                current_view, new_view;
  rgrl_converge_status_sptr                     current_status;

  rgrl_feature_set_sptr                         from_set;
  rgrl_feature_set_sptr                         to_set;
  rgrl_matcher_sptr                             matcher;
  rgrl_scale_estimator_unwgted_sptr             unwgted_scale_est;
  rgrl_scale_estimator_wgted_sptr               wgted_scale_est;
  rgrl_weighter_sptr                            weighter;
  rgrl_match_set_sptr                           match_set;
  vcl_vector<rgrl_estimator_sptr>               xform_estimators;
  bool                                          failed, scale_in_range;

  failed = false;
  scale_in_range = true;

  new_view = initial_view;

  do { // for each stage/resolution
    data_->get_data_at_stage( new_view->resolution(),
                              from_set,
                              to_set,
                              matcher,
                              weighter,
                              unwgted_scale_est,
                              wgted_scale_est,
                              xform_estimators);
    match_set = 0;

    TRACE( 1, " Current resolution "<< new_view->resolution() <<"\n" );
    
    // If no estimator found for the current stage, the default is
    // from the initializer. View_based can deal with multiple
    // estimators in each stage/resolution.
    //
    if ( xform_estimators.empty() ) {
      xform_estimators.push_back( new_view->xform_estimator() );
    }

    // Check if the weighter and scale estimators are set. If the
    // weighter is null, set it to the unit weighter, i.e. the weight
    // is set to 1 for every match. If the no scale estimators are
    // given, set the unwgted one to null, i.e., the sale is 1.
    //
    if ( !weighter )
      weighter = new rgrl_weighter_unit();
    if ( !unwgted_scale_est && !wgted_scale_est )
        unwgted_scale_est = new rgrl_scale_est_null();
    

    unsigned iterations_at_level = 0; //keeps track of total iter at stage
    unsigned icp_iter_at_level = 0;   //keeps track of the icp iter 
                                      //after regions converged

    current_status = 0;
    bool should_estimate_scale = true;
    int  scale_est_count = 0;

    do { // for each view 
      current_view = new_view;
      
      TRACE( 3, "Doing new view on region (" << current_view->region().x0() << " - " << current_view->region().x1() <<" )" << vcl_endl );
      
      TRACE( 3, " Computing matches and scales\n" );
      // Compute matches, and scales for each feature set.
      //          
      match_set = matcher->compute_matches( *from_set,
                                            *to_set,
                                            *current_view,
                                            *scale );
      TRACE( 3, "      Matches: " << match_set->from_size()<<" \n" );
      
      // For the first iteration, use prior scale or estimate the
      // scales without weights (since we don't have any...)
      //
      if ( iterations_for_scale_est_ >= 0 && 
           scale_est_count > iterations_for_scale_est_ ) {
        should_estimate_scale = false;
      }

      rgrl_scale_sptr  new_scale = 0;
      if ( !should_estimate_scale ) {
        TRACE( 3, "No scale estimation \n" );   
      }
      else if ( match_set->from_size() == 0 ) {
        vcl_cerr<<"rgrl_view_based_registration:: Empty match set!!!"<<vcl_endl;
        failed = true;
        continue;
      }
      else { //should_estimate_scale && match_set->from_size() > 0
        if( !wgted_scale_est ) {              
          new_scale = unwgted_scale_est->estimate_unweighted( *match_set, scale );
        } else {
          if( iterations_at_level == 0 && !scale ) {
            assert( unwgted_scale_est );
            new_scale = unwgted_scale_est->estimate_unweighted( *match_set, scale );
          }
          else {
            weighter->compute_weights(*scale, *match_set);
            new_scale = wgted_scale_est->estimate_weighted( *match_set, scale );
          }
        }
        
        // If the new scale exists, and the geometric scale is above
        // the lower bound, replace the old one by the new one
        if ( new_scale ) {
          TRACE( 3, "New geometric scale = "<<new_scale->geometric_scale()<<"\n" );
          if ( new_scale->has_geometric_scale() && 
               new_scale->geometric_scale() < expected_min_geometric_scale_ ) 
            should_estimate_scale = false;
          else {
            scale = new_scale;
            scale_est_count++;
          }
        }
      }
      assert( scale );
      

      // Transformation estimation for the set of available estimators
      //
      // Transformation estimation using a simplified irls
      // (Iterative-Reweighted Least-Squares) routine.
      //
      // The match sets and the scales are fixed, but the associated
      // weights are updated (hence reweighted-least-squares)
      // throughout the estimation.
      //
      TRACE( 3, " Computing transformations of available transformation models\n" );

      unsigned num_models = xform_estimators.size();
      unsigned curr_model = 0;
      vcl_vector< rgrl_transformation_sptr > xforms(num_models, 0);
      vcl_vector< double > alignment_errors(num_models, 0);
      for (  ; curr_model < num_models; ++curr_model ) {
        if ( xform_estimators[curr_model]->type() == current_view->xform_estimator()->type() ) break;
      }
      if ( curr_model == num_models )//curr_model is not in the model hierarchy
        curr_model = 0;
      
      for ( ; curr_model < num_models; ++curr_model ) {
        xforms[curr_model] = current_view->xform_estimate();
        if ( !rgrl_util_irls( match_set, scale, weighter,
                              *conv_tester_, xform_estimators[curr_model], 
                              xforms[curr_model], alignment_errors[curr_model]) )
          break; //early termination if the current model failed.
      }
      
      // Generate next view. This is where the region might grow, and
      // estimator might change, depending in the type of
      // view_generator.
      //
      new_view = 
        view_generator_->next_view( *current_view, match_set,
                                    scale, xforms, xform_estimators, 
                                    alignment_errors );
      if ( !new_view->is_valid() ) {
        failed = true;
        continue;
      }

      // For debugging ...
      //
      rgrl_set_of<rgrl_match_set_sptr> match_sets;
      match_sets.push_back( match_set );
      if( vb_callback ) vb_callback( new_view, match_sets );
      
      // Update the weights and scale estimates based on the new view
      //
      TRACE( 3, " Updating scale estimates and checking for validity\n" );
      
      match_set->remap_from_features( *new_view->xform_estimate() );
      weighter->compute_weights( *scale, *match_set );

      if ( should_estimate_scale ) {
        if( !wgted_scale_est )            
          scale = unwgted_scale_est->estimate_unweighted( *match_set, scale );
        else
          scale = wgted_scale_est->estimate_weighted( *match_set, scale );
        
        TRACE( 3, " Geometric scale after converged = "<< scale->geometric_scale()<<"\n" );
      }

      if ( expected_max_geometric_scale_ > 0 && 
           scale->has_geometric_scale() && 
           scale->geometric_scale() > expected_max_geometric_scale_) {
        scale_in_range = false;
        failed = true;
      }
      
      // Perform convergence test
      //
      current_status = 
        conv_tester_->compute_status( current_status,
                                      *current_view, *new_view,
                                      match_set, scale );
      TRACE( 4, "run: (icp_iter_at_level = " << icp_iter_at_level << "),(iterations_at_level = " << iterations_at_level << ") oscillation count = " << current_status->oscillation_count() << "\n" );
      TRACE( 4, "run: error = " << current_status->error() << vcl_endl );
      TRACE( 4, "run: error_diff = " << current_status->error_diff() << vcl_endl );

      // If the reigstration regions stop growing between views, the
      // registration becomes standard ICP. This will trigger
      // increment of icp_iter_at_level. If the regions grow again,
      // icp_iter_at_level is reset to 0.
      //
      if (new_view->regions_converged_to( *current_view ) ) {
        ++icp_iter_at_level;
        TRACE( 3, "icp iter = "<<icp_iter_at_level<<" \n" );
      } else {
        icp_iter_at_level = 0;
      }
      
      ++iterations_at_level;
      
    } while( !failed &&
             !current_status->has_converged() &&
             !current_status->has_stagnated() &&
             icp_iter_at_level < max_icp_iter_ );

    if ( failed ) {
        if ( !scale_in_range )
          TRACE( 3, " Geometric scale above the expected value \n" );
        else TRACE( 3, " Invalid view \n" );
        continue;
    }
    if ( current_status->has_converged() )
      TRACE( 3, " CONVERGED \n" );
    if ( current_status->has_stagnated() )
      TRACE( 3, " STAGNATED \n" );
    if ( icp_iter_at_level == max_icp_iter_ ) 
      TRACE( 3, " ICP iteration reached maximum ("<<max_icp_iter_<<" )\n" );
    
    TRACE( 3, " Done. \n" );
    
    // Move to the next resolution with proper initialization if not
    // at the finest level
    //
    current_view = new_view;
    new_view = initialize_view_for_next_level(*current_view);
    
  } while ( !failed && 
            ( !new_view->is_at_finest_resolution() ||  
              !current_view->is_at_finest_resolution() ) );
  
  TRACE( 3, "Estimation complete\n" );
  
  //   For views that have over-approximated global region 
  //   (larger than the actual size), the current region might not able
  //   to converge to the global region. To avoid such problem, the current
  //   fix is to re-approximate the maximum region using the entire 
  //   from-sets. However if the change is too drastic (see explanation
  //   in the function), the global region is intact.
  //
  if ( !failed && !current_view->current_region_converged() ) 
    current_view = verify_global_region(*current_view, from_set, to_set, 
                                        matcher, scale);
  
  //   At this point the iterations are over for this initial
  //   estimate.  If the estimate is good enough, if it is the
  //   first one tested or the current view reached the maximum and 
  //   it is the best seen thus far, record information about this estimate.
  //
  rgrl_transformation_type final_model = xform_estimators[xform_estimators.size()-1]->type();
  
  if ( !failed && current_view->current_region_converged() && 
       current_view->xform_estimate()->type() == final_model ){
    if ( !best_view_ ||
         current_status->objective_value() < best_status_->objective_value() ) {
      best_view_ = current_view;
      if (best_scales_.size() != 1)
        best_scales_.resize(1);
      best_scales_[0] = scale;
      if (best_matches_.size() != 1)
        best_matches_.resize(1);
      best_matches_[0] = match_set;
      best_status_ = current_status;
      TRACE( 3, "Set best view\n" );
    }
  }
}

void 
rgrl_view_based_registration::
register_multi_feature( rgrl_view_sptr           initial_view,
                        rgrl_scale_sptr          prior_scale )
{
  rgrl_view_sptr                                current_view, new_view;
  rgrl_converge_status_sptr                     current_status;

  vcl_vector<rgrl_feature_set_sptr>             from_sets;
  vcl_vector<rgrl_feature_set_sptr>             to_sets;
  vcl_vector<rgrl_matcher_sptr>                 matchers;
  vcl_vector<rgrl_scale_estimator_unwgted_sptr> unwgted_scale_ests;
  vcl_vector<rgrl_scale_estimator_wgted_sptr>   wgted_scale_ests;
  vcl_vector<rgrl_weighter_sptr>                weighters;
  rgrl_set_of<rgrl_match_set_sptr>              match_sets;
  rgrl_set_of<rgrl_scale_sptr>                  scales;
  vcl_vector<rgrl_estimator_sptr>               xform_estimators;
  bool                                          failed, scale_in_range, use_prior_scale;

  use_prior_scale = false;
  if ( prior_scale ) use_prior_scale = true;
  scale_in_range = true;
  failed = false;

  new_view = initial_view;

  do { // for each stage/resolution
    data_->get_data_at_stage( new_view->resolution(),
                              from_sets,
                              to_sets,
                              matchers,
                              weighters,
                              unwgted_scale_ests,
                              wgted_scale_ests,
                              xform_estimators);
    unsigned data_count = from_sets.size();
    
    TRACE( 1, " Current resolution "<< new_view->resolution() <<"\n" );

    // Initialize the size of match_sets and scales to be the same as
    // the from_sets
    //
    match_sets.clear();
    scales.clear();
    match_sets.resize( data_count );
    scales.resize( data_count );

    // If no estimator found for the current stage, the default is
    // from the initializer. View_based can deal with multiple
    // estimators in each stage/resolution.
    //
    if ( xform_estimators.empty() ) {
      xform_estimators.push_back( new_view->xform_estimator() );
    }

    // Check if the weighter and scale estimators are set. If the
    // weighter is null, set it to the unit weighter, i.e. the weight
    // is set to 1 for every match. If the no scale estimators are
    // given, set the unwgted one to null, i.e., the sale is 1.
    //
    for( unsigned int fs=0; fs < data_count; ++fs ) {
      if ( !weighters[fs] )
        weighters[fs] = new rgrl_weighter_unit();
      if ( !unwgted_scale_ests[fs] && !wgted_scale_ests[fs] )
        unwgted_scale_ests[fs] = new rgrl_scale_est_null();
    }
    
    // If the initial view comes with a prior scale, initialize the scales 
    // using the prior scale.
    if ( use_prior_scale ) {
      TRACE( 3, "Prior scale = "<<prior_scale->geometric_scale()<<"\n" );
      for (  unsigned int fs = 0; fs < data_count; ++fs )
        scales[fs] = prior_scale;
    }
    
    unsigned iterations_at_level = 0; //keeps track of total iter at level
    unsigned icp_iter_at_level = 0;   //keeps track of the icp iter 
                                      //after regions converged

    current_status = 0;
    bool should_estimate_scale = true;
    int  scale_est_count = 0;

    do { // for each view 
        current_view = new_view;

        TRACE( 3, "Doing new view on region (" << current_view->region().x0() << " - " << current_view->region().x1() <<" )" << vcl_endl );

        TRACE( 3, " Computing matches and scales\n" );
        // Compute matches, and scales for each feature set.
        //
        for( unsigned int fs=0; fs < data_count; ++fs ) {
          TRACE( 3, "   Data set " << fs << vcl_endl );
          rgrl_match_set_sptr new_matches =
            matchers[fs]->compute_matches( *from_sets[fs],
                                           *to_sets[fs],
                                           *current_view,
                                           *scales[fs] );
          TRACE( 3, "      Matches: " << new_matches->from_size() <<" ("<<new_matches<<")\n" );
          
          // For the first iteration,use prior scale or estimate the
          // scales without weights (since we don't have any...)
          //
          if ( iterations_for_scale_est_ >= 0 && 
               scale_est_count > iterations_for_scale_est_ ) {
            should_estimate_scale = false;
          }

          rgrl_scale_sptr new_scale = 0;
          if ( !should_estimate_scale ) {
            TRACE( 3, "No scale estimation \n" );
          }
          else if ( new_matches->from_size() == 0 ) {
            vcl_cerr<<"rgrl_view_based_registration:: Empty match set!!!"<<vcl_endl;
            failed = true;
            continue;
          }
          else { //should_estimate_scale && new_matches->from_size() > 0 
            if( !wgted_scale_ests[fs] ) {              
              new_scale = unwgted_scale_ests[fs]->estimate_unweighted( *new_matches, scales[fs] );
            } else {
              if( iterations_at_level == 0 && !scales[fs] ) {
                assert( unwgted_scale_ests[fs] );
                new_scale = unwgted_scale_ests[fs]->estimate_unweighted( *new_matches, scales[fs]);
              }
              else {
                weighters[fs]->compute_weights(*scales[fs], *new_matches);
                new_scale = wgted_scale_ests[fs]->estimate_weighted( *new_matches, scales[fs]);
              }
            }

            if ( new_scale ) {
              TRACE( 3, "New geometric scale = "<<new_scale->geometric_scale()<<"\n" );
              if ( new_scale->has_geometric_scale() && 
                   new_scale->geometric_scale() < expected_min_geometric_scale_ ) 
                should_estimate_scale = false;
              else {
                scale_est_count++;
                scales[fs] = new_scale;
              }
            }
          }
          assert( scales[fs] );

          // Keep new ones and discard old ones
          //
          match_sets[fs] = new_matches;
        }           

        TRACE( 3, " Matches[0]="<<match_sets[0]<<"\n" );
        TRACE( 3, " Computing transformations of available transformation models\n" );

        // Transformation estimation
        //
        unsigned num_models = xform_estimators.size();
        unsigned curr_model = 0;
        vcl_vector< rgrl_transformation_sptr > xforms(num_models, 0);
        vcl_vector< double > alignment_errors(num_models, 0);
        for (  ; curr_model < num_models; ++curr_model ) {
          if ( xform_estimators[curr_model]->type() == current_view->xform_estimator()->type() ) break;
        }
        if ( curr_model == num_models )//curr_model is not in the model hierarchy
          curr_model = 0;

        for ( ; curr_model < num_models; ++curr_model ) {
          xforms[curr_model] = current_view->xform_estimate();
          if ( !rgrl_util_irls( match_sets, scales, weighters,
                                *conv_tester_, xform_estimators[curr_model], 
                                xforms[curr_model], alignment_errors[curr_model]) )
            break; //early termination if the current model failed.
        }

        // Generate next view. This is where the region might grow, and
        // estimator might change, depending in the type of
        // view_generator.
        //
        new_view = 
          view_generator_->next_view( *current_view, match_sets,
                                      scales, xforms, xform_estimators, 
                                      alignment_errors );
        if ( !new_view->is_valid() ) {
          failed = true;
          continue;
        }

        // For debugging ...
        //
        if( vb_callback ) vb_callback( new_view, match_sets );

        // Update the weights and scale estimates based on the new transform
        //
        TRACE( 3, " Updating scale estimates and checking for validity\n" );
        for( unsigned int fs=0; fs < data_count; ++fs ) {
          if ( match_sets[fs]->from_size() > 0 ) {
            match_sets[fs]->remap_from_features( *new_view->xform_estimate() );
            weighters[fs]->compute_weights( *scales[fs], *match_sets[fs] );

            if ( should_estimate_scale ) {
              if( !wgted_scale_ests[fs] )            
                scales[fs] = unwgted_scale_ests[fs]->estimate_unweighted( *match_sets[fs], scales[fs] );
              else
                scales[fs] = wgted_scale_ests[fs]->estimate_weighted( *match_sets[fs], scales[fs] );
              
              TRACE( 3, " Geometric scale after converged = "<< scales[fs]->geometric_scale()<<"\n" );
            }

            if ( expected_max_geometric_scale_ > 0 && 
                 scales[fs]->has_geometric_scale() && 
                 scales[fs]->geometric_scale() > expected_max_geometric_scale_) {
              scale_in_range = false;
              failed = true;
            }
          }
        }

        // Perform convergence test
        //
        current_status = 
          conv_tester_->compute_status( current_status,
                                        *current_view, *new_view,
                                        match_sets, scales );
        TRACE( 4, "run: (icp_iter_at_level = " << icp_iter_at_level << "),(iterations_at_level = " << iterations_at_level << ") oscillation count = " << current_status->oscillation_count() << "\n" );
        TRACE( 4, "run: error = " << current_status->error() << vcl_endl );
        TRACE( 4, "run: error_diff = " << current_status->error_diff() << vcl_endl );
    
        // If the reigstration regions stop growing between views, the
        // registration becomes standard ICP. This will trigger
        // increment of icp_iter_at_level. If the regions grow again,
        // icp_iter_at_level is reset to 0.
        //
        if (new_view->regions_converged_to( *current_view ) ) {
          ++icp_iter_at_level;
          TRACE( 3, "icp iter = "<<icp_iter_at_level<<" \n" );
        } else {
          icp_iter_at_level = 0;
        }

        ++iterations_at_level;

      } while( !failed &&
               !current_status->has_converged() &&
               !current_status->has_stagnated() &&
               icp_iter_at_level < max_icp_iter_ );

      if ( failed ) {
        if ( !scale_in_range )
          TRACE( 3, " Geometric scale above the expected value \n" );
        else TRACE( 3, " Invalid view \n" );
        continue;
      }
      if ( current_status->has_converged() )
        TRACE( 3, " CONVERGED \n" );
      if ( current_status->has_stagnated() )
        TRACE( 3, " STAGNATED \n" );
      if ( icp_iter_at_level == max_icp_iter_ ) 
        TRACE( 3, " ICP iteration reached maximum ("<<max_icp_iter_<<" )\n" );
      
      TRACE( 3, " Done. \n" );
      
      // Move to the next resolution with proper initialization if not
      // at the finest level
      //
      current_view = new_view;
      new_view = initialize_view_for_next_level(*current_view);
      use_prior_scale = true;
      prior_scale = scales[0]; //Assuming scales[0] is a good approximate

    } while ( !failed && 
              ( !new_view->is_at_finest_resolution() ||  
                !current_view->is_at_finest_resolution() ) );
 
    TRACE( 3, "Estimation complete\n" );

    //   For views that have over-approximated global region 
    //   (larger than the actual size), the current region might not able
    //   to converge to the global region. To avoid such problem, the current
    //   fix is to re-approximate the maximum region using the entire 
    //   from-sets. However if the change is too drastic (see explanation
    //   in the function), the global region is intact.
    //
    if ( !failed && !current_view->current_region_converged() ) 
      current_view = verify_global_region(*current_view, from_sets, to_sets, 
                                          matchers, scales);

    //   At this point the iterations are over for this initial
    //   estimate.  If the estimate is good enough, if it is the
    //   first one tested or the current view reached the maximum and 
    //   it is the best seen thus far, record information about this estimate.
    //
    rgrl_transformation_type final_model = xform_estimators[xform_estimators.size()-1]->type();

    if ( !failed && current_view->current_region_converged() && 
         current_view->xform_estimate()->type() == final_model ){
      if ( !best_view_ ||
           current_status->objective_value() < best_status_->objective_value() ) {
        best_view_ = current_view;
        best_scales_ = scales;
        best_matches_ = match_sets;
        best_status_ = current_status;
        TRACE( 3, "Set best view\n" );
      }
    }
}

rgrl_view_sptr
rgrl_view_based_registration::
initialize_view_for_next_level( rgrl_view const&  current_view)
{
  // Find the next available level
  //
  unsigned new_resol = current_view.resolution();
  int next_resol = new_resol - 1;
  for ( ; next_resol >= 0; --next_resol) {
    if ( data_->has_stage(next_resol) ) { 
      new_resol = next_resol;
      break;
    }
  }

  // If no next available level return the same view
  //
  if ( new_resol == current_view.resolution() ) 
    return new rgrl_view( current_view );

  // Scale the components of the current view for the initial view of the next
  // available level
  //
  int level_diff = current_view.resolution() - new_resol;
  double dim_increase = data_->dimension_increase_for_next_stage(current_view.resolution());
  double scaling = vcl_pow(dim_increase, level_diff);

  rgrl_roi from_image_roi( current_view.from_image_roi().x0()*scaling,
                           current_view.from_image_roi().x1()*scaling );
  rgrl_roi to_image_roi( current_view.to_image_roi().x0()*scaling,
                         current_view.to_image_roi().x1()*scaling );
  rgrl_roi current_region( current_view.region().x0()*scaling,
                           current_view.region().x1()*scaling);
  rgrl_roi global_region( current_view.global_region().x0()*scaling,
                          current_view.global_region().x1()*scaling );
  rgrl_transformation_sptr  xform_estimate = current_view.xform_estimate()->scale_by( scaling );

  return new rgrl_view( from_image_roi, to_image_roi, 
                        current_region, global_region,
                        current_view.xform_estimator(),
                        xform_estimate,
                        new_resol);
}

rgrl_view_sptr 
rgrl_view_based_registration::
verify_global_region( rgrl_view const&                 curr_view,                      
                      rgrl_feature_set_sptr            from_set,
                      rgrl_feature_set_sptr            to_set,
                      rgrl_matcher_sptr                matcher,
                      rgrl_scale_sptr                  scale) const
{
  vcl_vector<rgrl_feature_set_sptr> from_sets;
  from_sets.push_back( from_set );
  vcl_vector<rgrl_feature_set_sptr> to_sets;
  from_sets.push_back( to_set );
  vcl_vector<rgrl_matcher_sptr> matchers;
  matchers.push_back( matcher );
  rgrl_set_of<rgrl_scale_sptr> scales;
  scales.push_back( scale );

  return this->verify_global_region( curr_view, from_sets, to_sets,
                                     matchers, scales);
}

rgrl_view_sptr
rgrl_view_based_registration::
verify_global_region( rgrl_view const&  curr_view,                      
                      vcl_vector<rgrl_feature_set_sptr> const& from_sets,
                      vcl_vector<rgrl_feature_set_sptr> const& to_sets,
                      vcl_vector<rgrl_matcher_sptr>     const& matchers,
                      rgrl_set_of<rgrl_scale_sptr>      const& scales) const
{
  TRACE( 3, "verify_global_region \n" );
  typedef rgrl_match_set::from_iterator from_iter;
  typedef from_iter::to_iterator        to_iter;

  rgrl_view_sptr temp_view = new rgrl_view(curr_view.from_image_roi(),
                                           curr_view.to_image_roi(),
                                           curr_view.global_region(),
                                           curr_view.global_region(),
                                           curr_view.xform_estimator(),
                                           curr_view.xform_estimate(),
                                           curr_view.resolution());
  rgrl_roi curr_global_region = curr_view.global_region(); 

  // Compute the global region as the minimum region enclosing all the 
  // points in the from_sets that mapped to the valid region of the 
  // to_sets.
  //
  vnl_vector<double> x0 = curr_view.from_image_roi().x1();
  vnl_vector<double> x1 = curr_view.from_image_roi().x0();
  unsigned int m = x0.size();
  unsigned data_count = from_sets.size();
  for( unsigned int fs=0; fs < data_count; ++fs ) {
    rgrl_match_set_sptr match_set =
      matchers[fs]->compute_matches( *from_sets[fs],
                                     *to_sets[fs],
                                     curr_view,
                                     *curr_view.xform_estimate(),
                                     *scales[fs] );
    for( from_iter fitr = match_set->from_begin();
         fitr != match_set->from_end(); ++fitr ) {
      if( fitr.size() == 0 )  continue;
      rgrl_feature_sptr from = fitr.from_feature();

      for (unsigned int i = 0; i<m; i++) {
        if (from->location()[i] < x0[i]) x0[i] = from->location()[i];
        if (from->location()[i] > x1[i]) x1[i] = from->location()[i];
      } 
    }
  }

  //Check if the new global region is valid
  //It is considered valid if lower right corner is greater than the upper 
  //left coner and the new ROI is not far too different from the current one
  //The measure of difference is somehow heuristic. If the region/volumn of
  //new ROI is n times larger or smaller than the current, then they are
  //considered different.
  bool valid = true;
  double space_new = 1, space_old = 1;
  for (unsigned int i = 0; i<m; i++) {
    if (x0[i] >= x1[i]) valid = false;
    space_new *= x1[i] - vnl_math_min(x0[i], x1[i]);
    space_old *= curr_global_region.x1()[i] - 
      vnl_math_min(curr_global_region.x0()[i], curr_global_region.x1()[i]);
  } 
  if (valid) { 
    double n = vcl_pow(5.0, (int)m);
    if (space_old/space_new > n ||
        space_new/space_old > n )
      valid = false;
  }

  // Check if the new global region is part of the current registration region 
  //
  bool in = true;
  rgrl_roi region = curr_view.region(); 
  for (unsigned int i = 0; i<m; i++) {
    if ( x0[i] < region.x0()[i] ) in = false;
    if ( x1[i] > region.x1()[i] ) in = false;
  } 
  TRACE( 3, "Estimated new global region = "<<x0<<", "<<x1<<"\n" );

  // If the new global region is part of the current region, it is safe to
  // set the new global region and the new registration region the same
  // for the corrected view.
  if ( valid && in ) {
    return new rgrl_view(curr_view.from_image_roi(),
                         curr_view.to_image_roi(),
                         rgrl_roi( x0, x1 ),
                         rgrl_roi( x0, x1 ),
                         curr_view.xform_estimator(),
                         curr_view.xform_estimate(),
                         curr_view.resolution());
  }
  
  TRACE( 3, "Global region intact = "<<curr_global_region.x0()<<", "<<curr_global_region.x1()<<"\n" );
  return new rgrl_view( curr_view );
}
