#ifndef rgrl_view_based_registration_h_
#define rgrl_view_based_registration_h_

//:
// \file
// \brief The engine for view-based registration. 
// \author Charlene Tsai
// \date   Feb 2004

#include "rgrl_data_manager_sptr.h"
#include "rgrl_convergence_tester_sptr.h"
#include "rgrl_view_generator_sptr.h"
#include "rgrl_estimator_sptr.h"
#include "rgrl_transformation_sptr.h"
#include "rgrl_set_of.h"
#include "rgrl_scale_sptr.h"
#include "rgrl_initializer_sptr.h"
#include "rgrl_converge_status_sptr.h"
#include "rgrl_match_set_sptr.h"
#include "rgrl_weighter_sptr.h"
#include "rgrl_roi.h"
#include "rgrl_view_sptr.h"
#include "rgrl_feature_set_sptr.h"
#include "rgrl_matcher_sptr.h"

//: View-based registration
//
//  The registration engine takes a data storage of single- or
//  multiple-stage/resolution. Each stage may contain multiple feature
//  types, and also multiple estimators. Depending on the \a
//  view_generator passed in the constructor, registration region
//  growth and model selection are possible with view-based
//  registration.
//
class rgrl_view_based_registration {
public:
  rgrl_view_based_registration( rgrl_data_manager_sptr        data,
                                rgrl_view_generator_sptr      view_generator,
                                rgrl_convergence_tester_sptr  conv_tester );
  
  ~rgrl_view_based_registration();

  //: Running from multiple initial estimates, produced by the initializer during registration
  //  Loop through the set of initial estimates.
  void run( rgrl_initializer_sptr initializer );

  //: Running from a given initial estimate. 
  //
  //  Based on if data_->is_multi_feature(), call run_single_feature(.)
  //  or run_multi_feature(.)
  //
  void run( rgrl_view_sptr            prior_view,
            rgrl_scale_sptr           prior_scale = 0 );


  //////////////// functions to access internal data  ////////////////////////

  //:  Return the final view for the best transformation estimate
  rgrl_view_sptr final_view() const;

  //:  Return true if has a best view
  bool has_best_view() const;

  //:  Return the final, best estimate
  rgrl_transformation_sptr final_transformation() const;

  //:  Return the scales of the best transformation estimate
  rgrl_set_of<rgrl_scale_sptr>  const& final_scales() const;
  rgrl_scale_sptr final_scale() const;
  
  //:  Return the status of the best transformation estimate.
  rgrl_converge_status_sptr final_status() const;

  //:  The multiple match sets used for the best transformation estimate.
  rgrl_set_of<rgrl_match_set_sptr> const& final_match_sets() const;

  //:  The single match set used for the best transformation estimate.
  rgrl_match_set_sptr final_match_set() const;

  //:  Return the number of initial transformations tested
  //
  //   The value might be greater than 1 if using initializer which
  //   provides a number of initial estiamtes.
  unsigned num_initial_xforms_tested() const;

  //: Set the max number of icp iteration per level
  //
  //  The default is 25 at initialization
  void set_max_icp_iter( unsigned iter );
  
  //: Set the expected max scale
  //
  //  If \a expected_max_geometric_scale_ is never set, the default
  //  has no effect on the registration.
  void set_expected_max_geometric_scale( double scale);

  //: Set the expected minimum geometric scale
  //
  //  If \a expected_min_geometric_scale_ is never set, the default
  //  has no effect on the registration.
  void set_expected_min_geometric_scale( double scale);

  //: Set the number of iterations during which the scale is to be estimated
  //
  //  If \a iterations_for_scale_est_ is never set, the default
  //  has no effect on the registration.
  void set_iterations_for_scale_est( int iter);


private:
  //: registration of single feature type at each stage/resolution
  void register_single_feature( rgrl_view_sptr           initial_view,
                                rgrl_scale_sptr          scale );

  //: registration of multiple feature type at each stage/resolution
  void register_multi_feature( rgrl_view_sptr           initial_view,
                               rgrl_scale_sptr          prior_scale );

  //: Scale/shrink the view according to the \a dimension_increase_for_next_stage at the current data stage
  rgrl_view_sptr 
  initialize_view_for_next_level( rgrl_view const&  current_view);

  //: Recompute the global region of the view using the match sets, instead of the corners of image regions
  rgrl_view_sptr 
  verify_global_region( rgrl_view const&  curr_view,                      
                        vcl_vector<rgrl_feature_set_sptr> const& from_sets,
                        vcl_vector<rgrl_feature_set_sptr> const& to_sets,
                        vcl_vector<rgrl_matcher_sptr>     const& matchers,
                        rgrl_set_of<rgrl_scale_sptr>      const& scales) const;

  //: Recompute the global region of the view using the match sets, instead of the corners of image regions
  rgrl_view_sptr 
  verify_global_region( rgrl_view const&                 curr_view,
                        rgrl_feature_set_sptr            from_set,
                        rgrl_feature_set_sptr            to_set,
                        rgrl_matcher_sptr                matcher,
                        rgrl_scale_sptr                  scales) const;

private:
  rgrl_data_manager_sptr data_;
  rgrl_convergence_tester_sptr conv_tester_;
  rgrl_view_generator_sptr view_generator_;

  rgrl_view_sptr                     best_view_;
  rgrl_converge_status_sptr          best_status_;
  rgrl_set_of<rgrl_match_set_sptr>   best_matches_;
  rgrl_set_of<rgrl_scale_sptr>       best_scales_;

  unsigned num_xforms_tested_;
  unsigned max_icp_iter_;
  double   expected_max_geometric_scale_; //for early termination
  double   expected_min_geometric_scale_;
  int      iterations_for_scale_est_; 
};

#endif
