//:
// \file
// \author Amitha Perera

#include "rgrl_view_gen_dbicp.h"

#include <vcl_cmath.h>
#include <vcl_cassert.h>
#include <vcl_iostream.h>
#include <vnl/vnl_vector.h>
#include <vnl/vnl_matrix.h>
#include <vnl/algo/vnl_determinant.h>
#include <vnl/vnl_math.h>

#include "rgrl_view.h"
#include "rgrl_roi.h"
#include "rgrl_transformation_sptr.h"
#include "rgrl_transformation.h"
#include "rgrl_scale.h"
#include "rgrl_util.h"
#include "rgrl_estimator.h"
#include "rgrl_match_set.h"
#include "rgrl_weighter.h"

// #include "rgrl_cast.h"
// #include "rgrl_trans_affine.h"

#if defined(_DEBUG) || defined(DEBUG)
#  include <vcl_iostream.h>
#  define DBG(x) x
#else
#  define DBG(x) /*no debug output*/
#endif

rgrl_view_gen_dbicp::
rgrl_view_gen_dbicp( double max_growth_rate, double min_growth_rate,
                     int max_num_update_global_region_per_model )
  : max_growth_rate_( max_growth_rate ),
    min_growth_rate_( min_growth_rate ),
    max_num_update_global_region_( max_num_update_global_region_per_model ),
    num_update_global_region_ (0 )
{
}


rgrl_view_sptr
rgrl_view_gen_dbicp::
next_view( rgrl_view const&                                curr_view,
           rgrl_set_of<rgrl_match_set_sptr> const&         curr_matches,
           rgrl_set_of<rgrl_scale_sptr> const&             curr_scales,
           vcl_vector< rgrl_transformation_sptr >  const&  xforms,
           vcl_vector<rgrl_estimator_sptr> const&          models,
           vcl_vector<double > const &                     errors )
{
  typedef vcl_vector<rgrl_estimator_sptr>::const_iterator model_itr;

  // This function performs the 3 operations in order
  // Step 1. update global region if the region catches up with the max region
  // Step 2. model selection
  // Step 3. region growing
  // 

  // Number of dimensions
  //
  unsigned const m = curr_view.region().x0().size();

  // Step 1. Update the global region if the current view reached the global region. 
  //
  rgrl_roi global_region(m);
  if ( curr_view.region() == curr_view.global_region() && 
       num_update_global_region_ < max_num_update_global_region_ ) { 
    global_region = rgrl_util_estimate_global_region(curr_view.from_image_roi(), 
                                                     curr_view.to_image_roi(),
                                                     curr_view.global_region(), 
                                                     *curr_view.xform_estimate(), 
                                                       true);
    DBG( vcl_cout<<"Updated global_region"<<vcl_endl );
    num_update_global_region_++;
  }
  else global_region = curr_view.global_region();
  DBG( vcl_cout<<"Global region = ( "<<global_region.x0()<<" - "<<global_region.x1()<<" )"<<vcl_endl );


  // Step 2. 
  // 1.  Model selection, which can happen if
  //     model selection criterion picked the higher-order model, or
  //     the region of the view reached the global, in which case, the next
  //     higher-order model is selected.
  //

  // start the estimation
  rgrl_estimator_sptr estimator;
  rgrl_transformation_sptr curr_xform;

  DBG( vcl_cout<<"Model selection"<<vcl_endl );
  unsigned int mi = 0;
  for (  ; mi!= models.size(); ++mi ) {
    if ( models[mi]->type() == curr_view.xform_estimator()->type() ) break;
  }

  bool best_model_set = false;
  double best_bmsc = 0;
  unsigned best_model = 0;

  for ( ; mi < models.size() && xforms[mi] ; ++mi ) {
    double stability = vcl_log(vnl_determinant(xforms[mi]->covar()));
    DBG( vcl_cout<<"Model type  "<<models[mi]->type()<<", "<<errors[mi]<<", "<<0.5*stability<<"\n");
    // bmsc = dof/2*log(2*pi) - sum(r^2_i*w_i) + log(det(cov(xform)))
    double bmsc = models[mi]->param_dof()/2*vcl_log(2*vnl_math::pi) - errors[mi] + 0.5*stability;

    if ( !best_model_set || best_bmsc < bmsc ) {
      best_model_set = true;
      best_bmsc = bmsc;
      best_model = mi;
    }
  }
  
  // If no model has valid xform, the view not changed 
  if ( !best_model_set ) {
    vcl_cout << " no model available \n" ;
    return new rgrl_view( curr_view.from_image_roi(),
                          curr_view.to_image_roi(),
                          curr_view.region(),
                          curr_view.global_region(),
                          0,
                          0,
                          curr_view.resolution() );
  }
  estimator = models[best_model];
  curr_xform = xforms[best_model];

  // 2. If the region reached the global, and model not changed, 
  //    force the model to the next higher order, if not already the highest.
  //
  if ( curr_view.region() == global_region && 
       estimator->type() == curr_view.xform_estimator()->type() ) {
    best_model++;
    if ( best_model <  models.size() )  {
      estimator =  models[best_model];
      curr_xform = xforms[best_model];
      DBG( vcl_cout<<"Global region reached. Model changed"<<vcl_endl );
    }
  }
  DBG( vcl_cout<<"New model type = "<<estimator->type()<<vcl_endl );
  
  if ( estimator->type() != curr_view.xform_estimator()->type() ) {
    num_update_global_region_ = 0;
  }
  
  // Step 3. Region growing
  //
  rgrl_roi const& region = curr_view.region();

  // Growth factor
  //
  //double const beta = 0.73205; // sqrt(3)-1
  //double const beta = 0.1;

  // Since cbrt not defined in vcl, beta is treated as in 2D
  double beta = vcl_sqrt(max_growth_rate_) - 1;

  // Center of the region
  //
  vnl_vector<double> centre = region.x0();
  centre += region.x1();
  centre /= 2.0;

  vnl_vector<double> new_x0 = region.x0();
  vnl_vector<double> new_x1 = region.x1();

  assert( curr_scales[0]->has_geometric_scale() );
//  double scale_estimate = curr_scales[0]->geometric_scale();

//   vcl_cout << "In gen_view dbicp:\n";
//   vcl_cout << "xform.covar=\n" << dynamic_cast<rgrl_trans_affine const&>(curr_xform).covar() << "\n";

  // For each axial direction
  //
  vnl_vector<double> normal( m, 0.0 );
  vnl_vector<double> p( centre );
  vnl_vector<double> xformed_normal( m );
  vnl_vector<double> xformed_p( m );
  vnl_matrix<double> trans_err_covar( m, m );
  for( unsigned d=0; d < m; ++d ) {
//     vcl_cout << "For d="<<d<<"\n";
    // Faces intersecting at x0
    {
      // center of face in direction d
      p[d] = region.x0()[d];
      double half_length = p[d] - centre[d];
      normal[d] = half_length < 0.0 ? -1.0 : 1.0;
      curr_xform->map_location( p, xformed_p );
      curr_xform->map_direction( p, normal, xformed_normal );
      trans_err_covar = curr_xform->transfer_error_covar( p );
      //trans_err_covar = scale_estimate * curr_xform.transfer_error_covar( p );
      //       vcl_cout << "p="<<p<<"\n";
      //      vcl_cout << "Xformed p="<<xformed_p<<"\n";
      //       vcl_cout << "Normal="<<normal<<"\n";
      //       vcl_cout << "Xformed normal="<<xformed_normal<<"\n";
      //       vcl_cout << "Xfer error covar=\n"<<trans_err_covar<<"\n";
      double sigma_s_sqr = dot_product( xformed_normal, trans_err_covar*xformed_normal );
//       vcl_cout << "sigma_s_sqr = " << sigma_s_sqr << "\n";
      // max(sigma_s_sqr,1)
      if( sigma_s_sqr < 1.0 ) {
        sigma_s_sqr = 1.0;
      }
      // the direction (add or subtract) is captured in the sign of half_length
      new_x0[d] += half_length * beta / sigma_s_sqr;
    }

    // Faces intersecting at x1
    {
      // center of face in direction d
      p[d] = region.x1()[d];
      double half_length = p[d] - centre[d];
      normal[d] = half_length < 0.0 ? -1.0 : 1.0;
      curr_xform->map_location( p, xformed_p );
      curr_xform->map_direction( p, normal, xformed_normal );
      trans_err_covar = curr_xform->transfer_error_covar( p );
      //trans_err_covar = scale_estimate * curr_xform.transfer_error_covar( p );
      double sigma_s_sqr = dot_product( xformed_normal, trans_err_covar*xformed_normal );
      if( sigma_s_sqr < 1.0 ) {
        sigma_s_sqr = 1.0;
      }
      // the direction (add or subtract) is captured in the sign of half_length
      new_x1[d] += half_length * beta / sigma_s_sqr;
    }

    // reset for next iteration
    p[d] = centre[d];
    normal[d] = 0.0;
  }

  DBG( vcl_cout << "Potential Growth x0 = " << new_x0 - region.x0() << "\n" );
  DBG( vcl_cout << "Potential Growth x1 = " << new_x1 - region.x1() << vcl_endl );

  // Check if the growth met the min growth rate. 
  // For now, do not swich model to the next available one. 
  //
  double prev_space = 1;
  double new_space = 1;
  vnl_vector<double>  prev_space_dim = region.x1() - region.x0();
  vnl_vector<double>  new_space_dim = new_x1 - new_x0;
  for( unsigned d=0; d < m; ++d ) {
    prev_space *= prev_space_dim[d];
    new_space  *= new_space_dim[d];
  }
  if ( new_space/prev_space < min_growth_rate_ ) { //restore the new region to the prev
    new_x0 = region.x0();
    new_x1 = region.x1();
    DBG( vcl_cout << "Low growth rate. View stops growing ...\n" );

    /*
    // force the model to the next available one
    //
    best_model++;
    if ( best_model <  models.size() )  {
      estimator =  models[best_model];
      curr_xform = xforms[best_model];
      DBG( vcl_cout<<"Model changed due to no region growth"<<vcl_endl );
    }
    */
  }

  // Clip at global ROI
  //
  vnl_vector<double> const& lo = global_region.x0();
  vnl_vector<double> const& hi = global_region.x1();
  for( unsigned d=0; d < m; ++d ) {
    assert( lo[d] <= hi[d] );
    if( new_x0[d] < lo[d] )  new_x0[d] = lo[d];
    if( new_x0[d] > hi[d] )  new_x0[d] = hi[d];
    if( new_x1[d] < lo[d] )  new_x1[d] = lo[d];
    if( new_x1[d] > hi[d] )  new_x1[d] = hi[d];
  }

  return new rgrl_view( curr_view.from_image_roi(),
                        curr_view.to_image_roi(),
                        rgrl_roi( new_x0, new_x1 ),
                        global_region,
                        estimator,
                        curr_xform,
                        curr_view.resolution() );
}
